from __future__ import annotations

import warnings
from collections.abc import Iterable
from dataclasses import asdict
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Type, Union

from pydantic import BaseModel, ValidationError, validator
from typing_extensions import TypedDict

from retentioneering.exceptions.widget import WidgetParseError
from retentioneering.params_model.registry import register_params_model
from retentioneering.utils.dict import clear_dict
from retentioneering.widget import WIDGET_MAPPING

if TYPE_CHECKING:
    from pydantic.typing import AbstractSetIntStr, MappingIntStrAny


class CustomWidgetProperties(TypedDict):
    widget: str
    serialize: Callable
    parse: Callable


class ParamsModel(BaseModel):
    _widgets: dict = {}

    @classmethod
    def __init_subclass__(cls: Type[ParamsModel], **kwargs: Any):
        super().__init_subclass__(**kwargs)
        register_params_model(cls)

    @validator("*")
    def validate_subiterable(cls, value: Any) -> Any:
        array_types = (Iterable, dict)
        if isinstance(value, array_types):
            try:
                if isinstance(value, dict):
                    subvalue = list(value.values())[0]
                else:
                    try:
                        subvalue = next(iter(value))
                    except StopIteration:
                        # empty iterator
                        subvalue = None
                if (isinstance(subvalue, array_types) or hasattr(subvalue, "__getitem__")) and not isinstance(
                    subvalue, str
                ):
                    raise ValueError("Inner iterable or hashable not allowed!")
            except TypeError:
                pass
        return value

    def __init__(
        self,
        **data: Dict[str, Any],
    ) -> None:
        try:
            super().__init__(**data)
        except ValidationError as v_error:
            for key in data:
                if key in self._widgets:
                    try:
                        data[key] = self._widgets[key]._parse(data[key])
                    except WidgetParseError as parse_err:
                        parse_err.field_name = key
                        raise parse_err
            super().__init__(**data)

    def __call__(self, **data: Dict[str, Any]) -> ParamsModel:
        ParamsModel.__init__(self, **data)
        return self

    @classmethod
    def _parse_schemas(cls) -> dict[str, str | dict | list]:
        with warnings.catch_warnings():
            # disable warning for pydantic schema Callable type
            warnings.simplefilter(action="ignore", category=UserWarning)
            params_schema: dict[str, Any] = cls.schema()

        params_schema["required"] = params_schema.get("required", [])

        for field_name, field in cls.__fields__.items():
            field_type = getattr(field, "type_", None)
            # TODO: python3.8 fix
            field_type_classname_legacy = getattr(field_type, "_name", None)
            field_type_classname = getattr(field_type, "__name__", None)

            if field.required and field_name not in params_schema["required"]:
                params_schema["required"].append(field_name)

            if field_type_classname == "Callable" or field_type_classname_legacy == "Callable":
                params_schema["properties"][field_name] = {
                    "title": field_name.title(),
                    "type": "callable",
                }
            elif not params_schema["properties"].get(field_name):
                params_schema["properties"][field_name] = {
                    "title": field_name.title(),
                    "type": "unknown",
                }
            params_schema["properties"][field_name]["default"] = field.default

        properties: dict[str, dict] = params_schema.get("properties", {})
        required: list[str] = params_schema.get("required", [])
        optionals = {name: name not in required for name in properties.keys()}
        definitions = params_schema.get("definitions", {})
        widgets = {}
        custom_widgets = cls._widgets
        for name, params in properties.items():
            custom_widget: dict[str, Any] | None = None
            default = params.get("default", None)
            if name in custom_widgets:  # type: ignore
                custom_widget = cls._parse_custom_widget(name=name, optional=optionals[name])

            if "$ref" in params or "allOf" in params:
                widget: dict[str, Any] = cls._parse_schema_definition(
                    params=params, definitions=definitions, default=default, optional=optionals[name]
                )
            elif params.get("type") == "string" and "enum" in params:
                widget = cls._parse_enum_schema_definition(
                    name=name, params=params, definitions=definitions, default=default, optional=optionals[name]
                )
            elif "anyOf" in params:
                widget = cls._parse_anyof_schema_definition(
                    params, definitions, default=default, optional=optionals[name]
                )
            else:
                widget = cls._parse_simple_widget(name, params, optional=optionals[name], default=default)

            if custom_widget:
                widget_optional = custom_widget.get("optional", widget.get("optional", False))
                widget_name = widget["name"] if "name" not in custom_widget else None
                widget_default = cls._get_serialized_default_value(
                    field_name=name,
                    widget_type=custom_widget.get("widget", None),
                    field_default=default,
                    widget_default=custom_widget.get("default", None),
                )
                custom_widget_data = {
                    "default": widget_default,
                    "optional": widget_optional,
                    "name": widget_name,
                }
                if isinstance(custom_widget_data["name"], str):
                    custom_widget_data["name"] = custom_widget_data["name"].lower().replace(" ", "_")

                custom_widget.update(clear_dict(custom_widget_data))
                widgets[name] = custom_widget
            elif widget:
                widget["default"] = cls._get_serialized_default_value(
                    field_name=name,
                    widget_type=widget.get("widget", None),
                    field_default=default,
                    widget_default=widget.get("default", None),
                )
                widgets[name] = widget
            else:
                raise ValueError("Not created widget")
        return widgets  # type: ignore

    @classmethod
    def _get_serialized_default_value(
        cls, field_name: str, widget_type: Optional[str] = None, field_default: Any = None, widget_default: Any = None
    ) -> Any:
        widget = WIDGET_MAPPING.get(widget_type, None) if widget_type else None
        default = widget_default if field_default is None else field_default

        # если дефолтного значения нет - значит его не нужно сериализовывать
        if default is None:
            return default

        serialize: None | Callable = None

        if field_name in cls._widgets:
            custom_widget = cls._widgets[field_name]  # type: ignore
            if isinstance(custom_widget, dict):
                serialize = custom_widget.get("_serialize", None)
            elif hasattr(custom_widget, "_serialize"):
                serialize = custom_widget._serialize

        if widget and hasattr(widget, "_serialize"):
            serialize = widget._serialize  # type: ignore

        if serialize is not None:
            return serialize(default)

        return default

    @classmethod
    def _parse_schema_definition(
        cls,
        params: dict[str, dict[str, Any]] | Any,
        definitions: dict[str, Any],
        default: Any | None = None,
        optional: bool = True,
    ) -> dict[str, Any]:
        ref: str = params.get("$ref", "") or params.get("allOf", [{}])[0].get("$ref", "")  # type: ignore
        definition_name = ref.split("/")[-1]
        definition = definitions[definition_name]
        params = definition.get("enum", [])
        kwargs = {"name": definition_name, "widget": "enum", "default": default, "optional": optional, "params": params}
        return asdict(WIDGET_MAPPING["enum"].from_dict(**kwargs))

    @classmethod
    def _parse_anyof_schema_definition(
        cls,
        params: dict[str, dict[str, Any] | list[dict[str, Any]]] | Any,
        definitions: dict[str, Any],
        default: Any | None = None,
        optional: bool = True,
    ) -> dict[str, Any]:
        definition_name = params.get("title")
        enum_params = [x.get("enum", [None])[0] for x in params["anyOf"] if hasattr(x, "get")]  # type: ignore
        enum_params = list(filter(lambda x: x is not None, enum_params))
        kwargs: dict = {"name": definition_name, "widget": "enum", "default": default, "optional": optional}
        if len(enum_params) > 0:
            kwargs["params"] = enum_params

        widget_data: dict[str, Any] = asdict(WIDGET_MAPPING["enum"].from_dict(**kwargs))
        kwargs.update(widget_data)
        return kwargs

    @classmethod
    def _parse_enum_schema_definition(
        cls,
        name: str,
        params: dict[str, dict[str, Any] | list[dict[str, Any]]] | Any,
        definitions: dict[str, Any],
        default: Any | None = None,
        optional: bool = True,
    ) -> dict[str, Any]:
        enum_params = params.get("enum")

        if not isinstance(enum_params, list):
            raise ValueError("unexpected enum value")

        enum_params = list(filter(lambda x: x is not None, enum_params))  # type: ignore
        kwargs: dict = {"name": name, "widget": "enum", "default": default, "optional": optional}
        if len(enum_params) > 0:
            kwargs["params"] = enum_params

        widget_data: dict[str, Any] = asdict(WIDGET_MAPPING["enum"].from_dict(**kwargs))
        kwargs.update(widget_data)
        return kwargs

    @classmethod
    def _parse_simple_widget(
        cls, name: str, params: dict[str, Any], default: Any | None = None, optional: bool = False
    ) -> dict[str, Any]:
        widget_type = params.get("type")
        if widget_type == "array":
            widget_type = "enum"
        try:
            items = params.get("items", [{}])[-1]
        except KeyError:
            items = params.get("items", [{}])
        widget_params = dict(optional=optional, name=name, widget=widget_type, default=default)

        if "enum" in items and widget_type != "enum":
            widget_type = "enum"
            widget_params["params"] = items["enum"]  # type: ignore

        return widget_params

    @classmethod
    def _parse_custom_widget(cls, name: str, optional: bool = False) -> dict[str, Any]:
        custom_widget = cls._widgets[name]  # type: ignore
        if isinstance(custom_widget, dict):
            _widget = WIDGET_MAPPING[custom_widget["widget"]]
            widget_type = custom_widget["widget"]
            return asdict(_widget.from_dict(**dict(optional=optional, name=name, widget=widget_type, value=None)))
        else:
            widget = asdict(custom_widget)
            if "optional" not in widget or widget["optional"] is None:
                widget["optional"] = optional
            return widget

    @classmethod
    def get_widgets(cls) -> dict[str, str | dict | list]:
        return cls._parse_schemas()

    def dict(
        self,
        *,
        include: Optional[Union["AbstractSetIntStr", "MappingIntStrAny"]] = None,
        exclude: Optional[Union["AbstractSetIntStr", "MappingIntStrAny"]] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> Dict:
        data = super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        for key in data:
            if data.get(key) is None:
                continue
            if widget := self._widgets.get(key, None):
                data[key] = widget._serialize(value=data[key])  # type: ignore
        return data



from typing import Callable, List, Optional, Tuple, Union, cast, get_args

import numpy as np
import pandas as pd
import plotly
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from sklearn.preprocessing import MinMaxScaler, StandardScaler

from retentioneering.common.constants import (
    PATH_METRIC_TYPES,
    SCALERS,
    SEGMENT_METRICS_OVERVIEW_KINDS,
)
from retentioneering.common.formatters import fancy_round
from retentioneering.eventstream.segments import get_segment_path_map
from retentioneering.eventstream.types import EventstreamType


class SegmentOverview:
    def __init__(self, eventstream: EventstreamType):
        self.__eventstream = eventstream
        self.segment_name: str
        self.table_1: pd.DataFrame
        self.table_2: pd.DataFrame
        self.table_1_annot: pd.DataFrame
        self.table_2_annot: pd.DataFrame
        self.min_value: float = 0.0
        self.max_value: float = 1.0
        self.heatmap_axis: int

    def _get_feature_table(
        self,
        features: pd.DataFrame,
        path_names: pd.Series,
        aggfunc: Union[Callable, str] = "mean",
        scaler: SCALERS = "minmax",
        axis: int = 1,
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        if scaler == "minmax":
            scaler_instance = MinMaxScaler()
        elif scaler == "std":
            scaler_instance = StandardScaler()
        else:
            raise ValueError(f"Unknown scaler {scaler}. Available scalers: {get_args(SCALERS)}")

        scaler_hm = MinMaxScaler(feature_range=(self.min_value, self.max_value))

        scaled_features = pd.DataFrame(
            scaler_instance.fit_transform(features), index=features.index, columns=features.columns
        )

        # hm stands for heatmap
        hm = scaled_features.groupby(path_names).agg(aggfunc)

        if axis == 0:
            hm = hm.transpose()

        hm = pd.DataFrame(scaler_hm.fit_transform(hm), index=hm.index, columns=hm.columns)

        if axis == 1:
            hm = hm.transpose()

        hm_annot = features.groupby(path_names).agg(aggfunc).transpose().pipe(fancy_round)

        return hm, hm_annot

    def _get_metrics_table(
        self,
        path_names: pd.Series,
        metrics: Optional[List[Tuple[PATH_METRIC_TYPES, Union[str, Callable], str]]],
        scaled: bool = True,
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        segment_size_col = "segment size"
        if metrics is None:
            metrics = [("segment_size", "mean", segment_size_col)]
        else:
            has_segment_size = False
            for metric in metrics:
                if metric[0] == "segment_size":
                    segment_size_col = metric[2]
                    has_segment_size = True
            if not has_segment_size:
                metrics.append(("segment_size", "mean", segment_size_col))

        scaler_cm = MinMaxScaler(feature_range=(self.min_value, self.max_value))
        # cm stands for custom path_metrics
        cm = pd.DataFrame()

        for metric_item in metrics:
            path_metric, segment_agg_metric, metric_name = metric_item
            if isinstance(path_metric, str) and path_metric == "segment_size":
                segment_agg_values = path_names.value_counts(normalize=True)
            else:
                path_metric = (path_metric, "tmp_metric_name")
                path_values = self.__eventstream.path_metrics(path_metric)
                segment_agg_values = path_values.groupby(path_names).agg(segment_agg_metric)
            cm[metric_name] = segment_agg_values

        cm = cm.sort_values(by=segment_size_col, ascending=False)
        cm_annot = cm.transpose().pipe(fancy_round)

        if scaled:
            cm = pd.DataFrame(scaler_cm.fit_transform(cm), index=cm.index, columns=cm.columns)

        cm = cm.transpose()

        return cm, cm_annot

    def fit_clusters_overview(
        self,
        segment_name: str,
        features: pd.DataFrame,
        aggfunc: Union[Callable, str] = "mean",
        scaler: SCALERS = "minmax",
        metrics: Optional[List[Tuple[PATH_METRIC_TYPES, Union[str, Callable], str]]] = None,
        axis: int = 1,
    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        path_names = cast(pd.Series, get_segment_path_map(self.__eventstream, name=segment_name))
        path_names = path_names.reindex(features.index).rename(segment_name)

        hm, hm_annot = self._get_feature_table(features, path_names, aggfunc, scaler, axis)
        cm, cm_annot = self._get_metrics_table(path_names, metrics)

        # sorting the columns by segment size
        hm = hm[cm.columns]
        hm_annot = hm_annot[cm.columns]

        # heatmap_axis
        if axis == 0:
            cm = pd.DataFrame(np.zeros(cm.values.shape) + 0.5, index=cm.index, columns=cm.columns)

        self.table_1 = hm
        self.table_2 = cm
        self.table_1_annot = hm_annot
        self.table_2_annot = cm_annot
        self.segment_name = segment_name
        self.axis = axis

        return hm, cm, hm_annot, cm_annot

    def fit_segment_overview(
        self,
        segment_name: str,
        metrics: Optional[List[Tuple[PATH_METRIC_TYPES, Union[str, Callable], str]]],
    ) -> pd.DataFrame:
        path_names = get_segment_path_map(self.__eventstream, name=segment_name, index="path_id")
        path_names = path_names.rename(segment_name)  # type: ignore

        cm, cm_annot = self._get_metrics_table(path_names, metrics, scaled=False)  # type: ignore

        self.table_2 = cm
        self.table_2_annot = cm_annot

        return cm

    def plot_heatmap(self) -> plotly.graph_objects.Figure:
        cell_height, cell_width = 20, 35
        character_width = 7
        colorbar_width = 100
        title_height = bottom_height = 100
        xtick_height = max(self.table_1.index.map(len)) * character_width
        index_width = self.table_1.index.union(self.table_2.index).str.len().max() * character_width

        width = index_width + cell_width * len(self.table_1.columns) + colorbar_width
        height = (
            cell_height * (len(self.table_1.index) + len(self.table_2.index))
            + title_height
            + bottom_height
            + xtick_height
        )

        hm_height_proportion = len(self.table_1.index) / (len(self.table_1.index) + len(self.table_2.index))
        cm_height_proportion = len(self.table_2.index) / (len(self.table_1.index) + len(self.table_2.index))

        fig = make_subplots(
            rows=2,
            cols=1,
            specs=[[{}], [{}]],
            row_heights=[hm_height_proportion, cm_height_proportion],
            vertical_spacing=0.01,
            shared_xaxes=True,
        )

        features_table = go.Heatmap(
            z=self.table_1,
            y=self.table_1.index,
            x=self.table_1.columns.astype(str),
            coloraxis="coloraxis1",
            xgap=1,
            ygap=1,
            text=self.table_1_annot,
            texttemplate="%{text}",
            hovertemplate=f"{self.segment_name}: %{{x}}<br>feature: %{{y}}<extra></extra>",
        )

        fig.add_trace(features_table, row=1, col=1)

        metrics_table = go.Heatmap(
            z=self.table_2,
            y=self.table_2.index,
            x=self.table_2.columns.astype(str),
            coloraxis="coloraxis2",
            xgap=1,
            ygap=1,
            text=self.table_2_annot,
            texttemplate="%{text}",
            hovertemplate=f"{self.segment_name}: %{{x}}<br>metric: %{{y}}<extra></extra>",
        )
        fig.add_trace(metrics_table, row=2, col=1)

        colorscale1 = "RdYlBu_r"
        colorscale2 = colorscale1 if self.axis == 1 else "Gray"
        tickangle = 0 if self.table_1.index.map(len).max() <= 3 else -90

        fig.update_layout(
            width=width,
            height=height,
            title_text=f"{self.segment_name} segment overview",
            title_x=0.5,
            coloraxis1=dict(
                colorscale=colorscale1,
                colorbar=dict(tickvals=[self.min_value, self.max_value], ticktext=["Low", "High"]),
            ),
            coloraxis2=dict(colorscale=colorscale2, showscale=False),
        )
        fig.update_xaxes(tickangle=tickangle)

        return fig

    def plot_custom_metrics(self, axis: int, kind: SEGMENT_METRICS_OVERVIEW_KINDS) -> plotly.graph_objects.Figure:
        cm = self.table_2
        cm_annot = self.table_2_annot

        if kind == "bar":
            if axis == 0:
                cm = cm.transpose()
            fig = px.bar(cm, barmode="group")

        elif kind == "heatmap":
            scaler = MinMaxScaler()
            if axis == 0:
                hm = cm.transpose()
            else:
                hm = cm.copy()

            # a workaround for timedelta columns
            # after transposing a timedelta columns loses its original dtype
            for col in hm.columns:
                try:
                    hm[col] = pd.to_timedelta(hm[col]).dt.total_seconds()
                except ValueError:
                    pass

            hm = pd.DataFrame(scaler.fit_transform(hm), index=hm.index, columns=hm.columns)

            if axis == 0:
                hm = hm.transpose()

            data = go.Heatmap(
                z=hm,
                y=hm.index,
                x=hm.columns.astype(str),
                coloraxis="coloraxis1",
                xgap=1,
                ygap=1,
                text=cm_annot,
                texttemplate="%{text}",
            )
            fig = go.Figure(data=data)

            fig.update_layout(
                coloraxis1=dict(
                    colorscale="RdYlBu_r",
                    colorbar=dict(tickvals=[self.min_value, self.max_value], ticktext=["Low", "High"]),
                )
            )
        else:
            raise ValueError(f"Unknown kind {kind}. Available kinds: {get_args(SEGMENT_METRICS_OVERVIEW_KINDS)}")

        return fig

    @staticmethod
    def _get_hover_values(df: pd.DataFrame) -> pd.DataFrame:
        new_data = []
        for i, row_index in enumerate(df.index):
            new_row = []
            for j, col_name in enumerate(df.columns):
                cell_value = df.iloc[i, j]
                new_row.append(f"feature: {row_index}<br>cluster_id: {col_name}<br>value: {cell_value}")
            new_data.append(new_row)
        new_df = pd.DataFrame(new_data, index=df.index, columns=df.columns)
        return new_df

    @property
    def values(self) -> Optional[Union[Tuple[pd.DataFrame, pd.DataFrame], pd.DataFrame]]:
        if hasattr(self, "table_1_annot") and hasattr(self, "table_2_annot"):
            return self.table_1_annot, self.table_2_annot
        elif not hasattr(self, "table_1_annot") and hasattr(self, "table_2_annot"):
            return self.table_2_annot
        return None



from __future__ import annotations

import math
from typing import Callable, Tuple

import numpy as np
import pandas as pd
import plotly.graph_objects as go
import seaborn as sns
from scipy.stats import chi2_contingency, fisher_exact, ks_2samp, mannwhitneyu
from scipy.stats.contingency import crosstab
from statsmodels.stats.power import TTestIndPower
from statsmodels.stats.weightstats import ttest_ind, ztest

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from retentioneering.eventstream.segments import _split_segment
from retentioneering.eventstream.types import (
    EventstreamType,
    SplitExpr,
    UserGroupsNamesType,
    UserGroupsType,
)
from retentioneering.tooling.stattests.constants import STATTEST_NAMES


def _effect_size_cohend(sample1: list, sample2: list) -> float:
    """
    Calculate an effect size value using Cohen’s D measure (difference between two means).

    Parameters
    ----------
    sample1 : list
        First group's statistics.

    sample2 : list
        Second group's statistics.

    Returns
    -------
    float
        Calculated an effect size value.

    """
    sample_size1, sample_size2 = len(sample1), len(sample2)
    standard_deviation1, standard_deviation2 = np.var(sample1, ddof=1), np.var(sample2, ddof=1)
    pooled_standard_deviation = float(
        math.sqrt(
            ((sample_size1 - 1) * standard_deviation1 + (sample_size2 - 1) * standard_deviation2)
            / (sample_size1 + sample_size2 - 2)
        )
    )
    mean_sample1, mean_sample2 = float(np.mean(sample1)), float(np.mean(sample2))
    return (mean_sample1 - mean_sample2) / pooled_standard_deviation


def _effect_size_cohenh(sample1: list, sample2: list) -> float:
    """
    Calculate an effect size value using Cohen’s D measure (between proportions)

    Parameters
    ----------
    sample1 : list
        First group's statistics.

    sample2 : list
        Second group's statistics.

    Returns
    -------
    float
        Calculated an effect size value.

    """
    mean_sample1, mean_sample2 = np.mean(sample1), np.mean(sample2)
    return 2 * (math.asin(math.sqrt(mean_sample1)) - math.asin(math.sqrt(mean_sample2)))


class StatTests:
    """
    A class for determining statistical difference between two groups of users.

    Parameters
    ----------
    eventstream : EventstreamType

    See Also
    --------
    .Eventstream.stattests : Call StatTests tool as an eventstream method.

    Notes
    -----
    See :doc:`StatTests user guide</user_guides/stattests>` for the details.

    """

    __eventstream: EventstreamType
    test: STATTEST_NAMES
    groups: UserGroupsType
    func: Callable
    group_names: UserGroupsNamesType
    alpha: float
    g1_data: list[str | int]
    g2_data: list[str | int]

    is_fitted: bool
    output_template_numerical = "{0} (mean ± SD): {1:.3f} ± {2:.3f}, n = {3}"
    output_template_categorical = "{0} (size): n = {1}"
    p_val, power, label_min, label_max = 0.0, 0.0, "", ""

    @time_performance(
        scope="stattests",
        event_name="init",
    )
    def __init__(self, eventstream: EventstreamType) -> None:
        self.__eventstream = eventstream
        self.user_col = self.__eventstream.schema.user_id
        self.event_col = self.__eventstream.schema.event_name
        self.time_col = self.__eventstream.schema.event_timestamp

        self.g1_data = list()
        self.g2_data = list()
        self.is_fitted = False

    def _get_group_values(self) -> Tuple[list, list]:
        data = self.__eventstream.to_dataframe()
        # obtain two populations for each group
        g1 = data[data[self.user_col].isin(self.groups[0])].copy()
        g2 = data[data[self.user_col].isin(self.groups[1])].copy()

        # obtain two distributions:
        if self.test not in ["chi2_contingency", "fisher_exact"]:
            g1_data = list(g1.groupby(self.user_col).apply(self.func).dropna().astype(float).values)
            g2_data = list(g2.groupby(self.user_col).apply(self.func).dropna().astype(float).values)
        else:
            g1_data = list(g1.groupby(self.user_col).apply(self.func).dropna().values)
            g2_data = list(g2.groupby(self.user_col).apply(self.func).dropna().values)
        return g1_data, g2_data

    def _get_freq_table(self, a: list, b: list) -> list:
        labels = ["A"] * len(a) + ["B"] * len(b)
        values = np.concatenate([a, b])
        if self.test == "fisher_exact" and np.unique(values).shape[0] != 2:
            raise ValueError("For Fisher exact test, there should be exactly 2 categories of observations in the data")
        return crosstab(labels, values)[1]

    def _get_test_results(self, data_max: list, data_min: list) -> Tuple[float, float]:
        if self.test in ["ztest", "ttest", "mannwhitneyu", "ks_2samp"]:
            # calculate effect size
            if max(data_max) <= 1 and min(data_max) >= 0 and max(data_min) <= 1 and min(data_min) >= 0:
                # if analyze proportions use Cohen's h:
                effect_size = _effect_size_cohenh(data_max, data_min)
            else:
                # for other variables use Cohen's d:
                effect_size = _effect_size_cohend(data_max, data_min)

            # calculate power
            power = TTestIndPower().power(
                effect_size=effect_size,
                nobs1=len(data_max),
                ratio=len(data_min) / len(data_max),
                alpha=self.alpha,
                alternative="larger",
            )

            if self.test == "ks_2samp":
                p_val = ks_2samp(data_max, data_min, alternative="less")[1]
            elif self.test == "mannwhitneyu":
                p_val = mannwhitneyu(data_max, data_min, alternative="greater")[1]
            elif self.test == "ttest":
                p_val = ttest_ind(data_max, data_min, alternative="larger")[1]
            elif self.test == "ztest":
                p_val = ztest(data_max, data_min, alternative="larger")[1]
        elif self.test in ["chi2_contingency", "fisher_exact"]:
            power = None
            if self.test == "chi2_contingency":
                freq_table = self._get_freq_table(data_max, data_min)
                p_val = chi2_contingency(freq_table)[1]
            elif self.test == "fisher_exact":
                freq_table = self._get_freq_table(data_max, data_min)
                p_val = fisher_exact(freq_table, alternative="greater")[1]
        else:
            raise ValueError(f"The argument test is not supported. Supported tests are: {STATTEST_NAMES}")
        return p_val, power  # type: ignore

    def _get_sorted_test_results(self) -> Tuple[float, float, str, str]:
        p_val_norm, power_norm = self._get_test_results(self.g1_data, self.g2_data)
        p_val_rev, power_rev = self._get_test_results(self.g2_data, self.g1_data)
        if p_val_norm < p_val_rev:
            p_val = p_val_norm
            power = power_norm
            label_max = self.group_names[0]
            label_min = self.group_names[1]
        else:
            p_val = p_val_rev
            power = power_rev
            label_max = self.group_names[1]
            label_min = self.group_names[0]
        return p_val, power, label_min, label_max

    @time_performance(
        scope="stattests",
        event_name="fit",
    )
    def fit(
        self,
        test: STATTEST_NAMES,
        groups: SplitExpr | UserGroupsType | str,
        func: Callable,
        group_names: UserGroupsNamesType = ("group_1", "group_2"),
        alpha: float = 0.05,
    ) -> None:
        """
        Calculates the stattests internal values with the defined parameters.
        Applying ``fit`` method is necessary for the following usage
        of any visualization or descriptive ``StatTests`` methods.

        Parameters
        ----------
        test : {'mannwhitneyu', 'ttest', 'ztest', 'ks_2samp', 'chi2_contingency', 'fisher_exact'}
            Test the null hypothesis that 2 independent samples are drawn from the same
            distribution. Supported tests are:

            - ``mannwhitneyu`` see :mannwhitneyu:`scipy documentation<>`.
            - ``ttest`` see :statsmodel_ttest:`statsmodels documentation<>`.
            - ``ztest`` see :statsmodel_ztest:`statsmodels documentation<>`.
            - ``ks_2samp`` see :scipy_ks:`scipy documentation<>`.
            - ``chi2_contingency`` see :scipy_chi2:`scipy documentation<>`.
            - ``fisher_exact`` see :scipy_fisher:`scipy documentation<>`.

        groups : tuple of list
            Must contain a tuple of two elements (g_1, g_2): g_1 and g_2 are collections
            of user_id`s.
        func : Callable
            Selected path_metrics. Must contain a function that takes a dataset as an argument for
            a single user trajectory and returns a single numerical value.
        group_names : tuple, default ('group_1', 'group_2')
            Names for selected groups g_1 and g_2.
        alpha : float, default 0.05
            Selected level of significance.

        """
        called_params = {
            "test": test,
            "groups": groups,
            "func": func,
            "group_names": group_names,
            "alpha": alpha,
        }
        not_hash_values = ["test"]

        self.groups, _ = _split_segment(self.__eventstream, groups)
        self.func = func
        self.test = test
        self.group_names = group_names
        self.alpha = alpha

        self.g1_data, self.g2_data = self._get_group_values()
        self.p_val, self.power, self.label_min, self.label_max = self._get_sorted_test_results()
        self.is_fitted = True
        collect_data_performance(
            scope="stattests",
            event_name="metadata",
            called_params=called_params,
            not_hash_values=not_hash_values,
            performance_data={},
            eventstream_index=self.__eventstream._eventstream_index,
        )

    @time_performance(
        scope="stattests",
        event_name="plot",
    )
    def plot(self) -> Tuple[go.Figure, str]:
        """
        Plots a barplot comparing the metric values between two groups.
        Should be used after :py:func:`fit`.

        Returns
        -------
        go.Figure
        """
        data1 = pd.DataFrame(data={"data": self.g1_data, "groups": self.group_names[0]})
        data2 = pd.DataFrame(data={"data": self.g2_data, "groups": self.group_names[1]})
        combined_stats = pd.concat([data1, data2]).reset_index()
        compare_plot = sns.displot(data=combined_stats, x="data", hue="groups", multiple="dodge")
        compare_plot.set(xlabel=None)
        return compare_plot

    @property
    @time_performance(
        scope="stattests",
        event_name="values",
    )
    def values(self) -> dict:
        """
        Returns the comprehensive results of the comparison between the two groups.
        Should be used after :py:func:`fit`.

        Returns
        -------
        dict

        """
        if not self.is_fitted:
            raise ValueError("The StatTests instance needs to be fitted before returning values")
        if self.test in ["ztest", "ttest", "mannwhitneyu", "ks_2samp"]:
            res_dict = {
                "group_one_name": self.group_names[0],
                "group_one_size": len(self.g1_data),
                "group_one_mean": np.array(self.g1_data).mean(),
                "group_one_SD": np.array(self.g1_data).std(),
                "group_two_name": self.group_names[1],
                "group_two_size": len(self.g2_data),
                "group_two_mean": np.array(self.g2_data).mean(),
                "group_two_SD": np.array(self.g2_data).std(),
                "greatest_group_name": self.label_max,
                "least_group_name": self.label_min,
                "is_group_one_greatest": self.label_max == self.group_names[0],
                "p_val": self.p_val,
                "power_estimated": self.power,
            }
        elif self.test in ["chi2_contingency", "fisher_exact"]:
            res_dict = {
                "group_one_name": self.group_names[0],
                "group_one_size": len(self.g1_data),
                "group_two_name": self.group_names[1],
                "group_two_size": len(self.g2_data),
                "p_val": self.p_val,
            }
        else:
            raise ValueError("Wrong test passed")
        return res_dict

    @time_performance(  # type: ignore
        scope="stattests",
        event_name="display_results",
    )
    def display_results(self) -> None:
        if not self.is_fitted:
            raise ValueError("The StatTests instance needs to be fitted before displaying results")
        values = self.values
        if self.test in ["ztest", "ttest", "mannwhitneyu", "ks_2samp"]:
            print(
                self.output_template_numerical.format(
                    values["group_one_name"], values["group_one_mean"], values["group_one_SD"], values["group_one_size"]
                )
            )
            print(
                self.output_template_numerical.format(
                    values["group_two_name"], values["group_two_mean"], values["group_two_SD"], values["group_two_size"]
                )
            )
            print(
                "'{0}' is greater than '{1}' with p-value: {2:.5f}".format(
                    values["greatest_group_name"], values["least_group_name"], values["p_val"]
                )
            )
            print("power of the test: {0:.2f}%".format(100 * values["power_estimated"]))
        elif self.test in ["chi2_contingency", "fisher_exact"]:
            print(self.output_template_categorical.format(values["group_one_name"], values["group_one_size"]))
            print(self.output_template_categorical.format(values["group_two_name"], values["group_two_size"]))
            print("Group difference test with p-value: {:.5f}".format(values["p_val"]))
        else:
            raise ValueError("Wrong test passed")

    @property
    @time_performance(
        scope="stattests",
        event_name="params",
    )
    def params(self) -> dict:
        """
        Returns the parameters used for the last fitting.
        Should be used after :py:func:`fit`.

        """
        return {
            "test": self.test,
            "groups": self.groups,
            "func": self.func,
            "group_names": self.group_names,
            "alpha": self.alpha,
        }



from __future__ import annotations

import base64
import functools
import hashlib
import inspect
import os
import uuid
from datetime import datetime
from typing import Any, Callable, Collection, Final

import pandas as pd

from retentioneering import RETE_CONFIG
from retentioneering.utils.hash_object import hash_value
from retentioneering.utils.singleton import Singleton

from ...utils.flatten_list import flatten
from .connector import ConnectorProtocol
from .tracking_info import TrackingInfo


class Tracker(Singleton):
    connector: ConnectorProtocol
    _enabled: bool = True
    session_id: str
    _user_id: str | None = None
    DO_NOT_TRACK: Final = "DO_NOT_TRACK"

    def __init__(self, connector: ConnectorProtocol, enabled: bool = True) -> None:
        self.connector = connector
        self._enabled = enabled
        # generate session id uuid
        self.session_id = str(uuid.uuid4())
        self.is_colab = 1 if self.check_env() == "colab" else 0

        self._track_action(
            called_function_params={},
            scope="tracker",
            event_name="init_tracker",
            event_custom_name="init_tracker",
            suffix="",
        )

    @staticmethod
    def get_top_call_id() -> str:
        top_call = [t for t in inspect.stack() if "tracker.py" in t.filename and t.function == "wrapper"][-1]

        top_frame_id = str(id(top_call.frame))
        top_frame_ts = str(top_call.frame.f_locals.get("start_time"))
        hasher = hashlib.sha1((top_frame_id + top_frame_ts).encode("utf-8"))
        return str(base64.urlsafe_b64encode(hasher.digest()[:10]))

    @staticmethod
    def check_env() -> str:
        try:
            import google.colab  # type: ignore # noqa: F401

            return "colab"
        except ImportError:
            return "classic"

    @property
    def enabled(self) -> bool:
        environment_tracker_status = os.environ.get("RETE_TRACKER_ENABLED", "true").lower()
        if environment_tracker_status not in ["true", "false"]:
            return True
        self._enabled = True if environment_tracker_status == "true" else False
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool | Any) -> None:
        self._enabled = value if isinstance(value, bool) else True

    @property
    def user_id(self) -> str:
        if self._user_id is None:
            self._user_id = self.__obtain_user_id()
        return self._user_id

    def __obtain_user_id(self) -> str:
        return RETE_CONFIG.user.pk

    def clear_params(self, params: dict[str, Any], not_hash_values: list[str] | None = None) -> dict[str, Any]:
        named_only_params = {key: self.DO_NOT_TRACK for key in params.keys()}
        value_allowed_params = self.__obtain_values(params=params, not_hash_values=not_hash_values)
        created_params = {**named_only_params, **value_allowed_params}
        return created_params

    @staticmethod
    def __prepare_value(value: Any, not_hash: bool) -> Any:
        if isinstance(value, (int, float, type(None))):
            return value
        elif isinstance(value, str):
            if not_hash:
                return value
            return hash_value(value)
        elif isinstance(value, pd.DataFrame):
            return value.shape
        elif isinstance(value, Collection):
            if not_hash:
                return value
            return {"len": len(value), "len_flatten": len(flatten(value))}
        elif callable(value):
            try:
                function_source = inspect.getsource(value)
            except Exception:
                function_source = ""
            return hash_value(function_source)
        return f"WRONG_TYPE: {type(value)}"

    def __obtain_values(
        self,
        params: dict[str, Any],
        not_hash_values: list[str] | None = None,
        hash_keys: bool = False,
    ) -> dict[str, Any]:
        prepared_params: dict[str, Any] = {}
        if not_hash_values is None:
            not_hash_values = []
        for key, value in params.items():
            tracking_key = hash_value(key) if hash_keys is True else key

            if isinstance(value, dict):
                prepared_params[tracking_key] = self.__obtain_values(
                    params=value,
                    hash_keys=False if key in not_hash_values else True,
                )
            else:
                prepared_params[tracking_key] = self.__prepare_value(value, key in not_hash_values)

        return prepared_params

    def _track_action(
        self,
        called_function_params: dict[str, Any],
        scope: str,
        event_name: str,
        event_custom_name: str,
        suffix: str,
        event_value: str = "",
        event_time: datetime | None = None,
        eventstream_index: int | None = None,
        parent_eventstream_index: int | None = None,
        child_eventstream_index: int | None = None,
        top_call_id: str = "",
    ) -> None:
        if self.enabled:
            try:
                if event_time is None:
                    event_time = datetime.now()
                tracking_info = TrackingInfo(
                    user_id=self.user_id,
                    event_name=event_name,
                    event_custom_name="_".join([str(_) for _ in [event_custom_name, suffix] if _]),
                    event_value=event_value,
                    params=called_function_params,
                    event_time=event_time,
                    scope=scope,
                    eventstream_index=eventstream_index,
                    jupyter_kernel_id=self.session_id,
                    parent_eventstream_index=parent_eventstream_index,
                    child_eventstream_index=child_eventstream_index,
                    top_call_id=top_call_id,
                    colab=self.is_colab,
                )
                self.connector.send_message(data=tracking_info)
            except Exception as e:
                # Maximum suppression. Vladimir Makhanov
                pass
        else:
            pass

    def __obtaion_evenstream_info(self, kwargs: dict[str, Any]) -> tuple[int | None, str | None]:
        if eventstream := kwargs.get("eventstream", None):
            return eventstream._eventstream_index, eventstream._hash
        return None, None

    def track(
        self,
        tracking_info: dict[str, Any],
        scope: str,
        event_value: str = "",
    ) -> Callable:
        event_name = tracking_info["event_name"]
        event_custom_name = tracking_info.get("event_custom_name", event_name)

        def tracker_decorator(func: Callable) -> Callable:
            @functools.wraps(func)
            def wrapper(*args: list[Any], **kwargs: dict[Any, Any]) -> Callable:
                called_function_params = self.clear_params(params=kwargs)
                eventstream_index, eventstream_hash = self.__obtaion_evenstream_info(kwargs)
                called_function_params["eventstream_index"] = eventstream_index
                called_function_params["eventstream_hash"] = eventstream_hash
                start_time = datetime.now()
                self._track_action(
                    called_function_params=called_function_params,
                    scope=scope,
                    event_name=event_name,
                    event_custom_name=event_custom_name,
                    event_value=event_value,
                    event_time=start_time,
                    suffix="start",
                    eventstream_index=eventstream_index,
                )

                try:
                    res = func(*args, **kwargs)
                except Exception as e:
                    raise e

                end_time = datetime.now()

                self._track_action(
                    called_function_params=called_function_params,
                    scope=scope,
                    event_name=event_name,
                    event_custom_name=event_custom_name,
                    event_value=event_value,
                    event_time=end_time,
                    suffix="end",
                    eventstream_index=eventstream_index,
                )
                return res

            return wrapper

        return tracker_decorator

    def time_performance(
        self,
        scope: str,
        event_name: str,
        event_value: str = "",
    ) -> Callable:
        def time_performance_decorator(func: Callable) -> Callable:
            @functools.wraps(func)
            def wrapper(*args: list[Any], **kwargs: dict[Any, Any]) -> Callable:
                eventstream_index, eventstream_hash = self.__obtaion_evenstream_info(kwargs)

                start_time = datetime.now()
                top_call_id = self.get_top_call_id()
                self._track_action(
                    called_function_params={},
                    scope=scope,
                    event_name=event_name,
                    event_custom_name=event_name,
                    event_value=event_value,
                    event_time=start_time,
                    suffix="start",
                    eventstream_index=eventstream_index,
                    top_call_id=top_call_id,
                )
                res = func(*args, **kwargs)
                end_time = datetime.now()
                self._track_action(
                    called_function_params={},
                    scope=scope,
                    event_name=event_name,
                    event_custom_name=event_name,
                    event_value=event_value,
                    event_time=end_time,
                    suffix="end",
                    eventstream_index=eventstream_index,
                    top_call_id=top_call_id,
                )

                return res

            return wrapper

        return time_performance_decorator

    def collect_data_performance(
        self,
        scope: str,
        event_name: str = "metadata",
        called_params: dict[str, Any] | None = None,
        performance_data: dict[str, Any] | None = None,
        event_value: str = "",
        not_hash_values: list[str] | None = None,
        eventstream_index: int | None = None,
        parent_eventstream_index: int | None = None,
        child_eventstream_index: int | None = None,
    ) -> None:
        if called_params is None:
            called_params = {}

        if performance_data is None:
            performance_data = {}

        called_function_params = self.clear_params(params=called_params, not_hash_values=not_hash_values)
        self._track_action(
            called_function_params={"args": called_function_params, "performance_info": performance_data},
            scope=scope,
            event_name=event_name,
            event_custom_name=event_name,
            event_value=event_value,
            suffix="",
            eventstream_index=eventstream_index,
            parent_eventstream_index=parent_eventstream_index,
            child_eventstream_index=child_eventstream_index,
        )



from __future__ import annotations

from typing import Any, Callable, Optional

import pandas as pd

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from retentioneering.utils.doc_substitution import docstrings

from ..types import EventstreamSchemaType, EventstreamType

EventstreamFilter = Callable[[pd.DataFrame, Optional[EventstreamSchemaType]], Any]


class GroupEventsHelperMixin:
    @docstrings.with_indent(12)
    @time_performance(  # type: ignore
        scope="group_events",
        event_name="helper",
        event_value="combine",
    )
    def group_events(
        self: EventstreamType,
        event_name: str,
        func: EventstreamFilter,
        event_type: str | None = "group_alias",
    ) -> EventstreamType:
        """
        A method of ``Eventstream`` class that filters and replaces raw events with new synthetic events,
        having the same ``timestamp`` and ``user_id``, but new ``event_name``.

        Parameters
        ----------
            %(GroupEvents.parameters)s

        Returns
        -------
        Eventstream
             Input ``eventstream`` with replaced events.



        """

        calling_params = {
            "event_name": event_name,
            "func": func,
            "event_type": event_type,
        }

        # avoid circular import
        from retentioneering.data_processors_lib import GroupEvents, GroupEventsParams
        from retentioneering.preprocessing_graph import PreprocessingGraph
        from retentioneering.preprocessing_graph.nodes import EventsNode

        p = PreprocessingGraph(source_stream=self)  # type: ignore

        node = EventsNode(
            processor=GroupEvents(
                params=GroupEventsParams(event_name=event_name, func=func, event_type=event_type)  # type: ignore
            )
        )
        p.add_node(node=node, parents=[p.root])
        result = p.combine(node)
        del p
        collect_data_performance(
            scope="group_events",
            event_name="metadata",
            called_params=calling_params,
            performance_data={},
            eventstream_index=self._eventstream_index,
            parent_eventstream_index=self._eventstream_index,
            child_eventstream_index=result._eventstream_index,
        )

        return result



from __future__ import annotations

from typing import Dict, Literal, TypedDict, Union

from typing_extensions import NotRequired, TypeAlias

from .normalization import NormalizationType

ColumnId: TypeAlias = str

ColumnNumericMap: TypeAlias = Dict[ColumnId, Union[float, int]]

ColumnValuesType: TypeAlias = Literal["number", "seconds"]


class Column(TypedDict):
    id: str
    name: str
    ## TODO: Implement this logic
    valuesType: NotRequired[ColumnValuesType]
    incompatibleNormalizationType: NotRequired[NormalizationType]
    description: NotRequired[str]



from __future__ import annotations

from typing import Optional, Tuple

import numpy as np
import pandas as pd

from retentioneering.backend.tracker import collect_data_performance, time_performance
from retentioneering.common.constants import DATETIME_UNITS
from retentioneering.data_processor import DataProcessor
from retentioneering.eventstream.types import EventstreamSchemaType
from retentioneering.params_model import ParamsModel
from retentioneering.utils.doc_substitution import docstrings
from retentioneering.utils.hash_object import hash_dataframe
from retentioneering.widget.widgets import ReteTimeWidget


class DropPathsParams(ParamsModel):
    """
    A class with parameters for :py:class:`.DropPaths` class.
    """

    min_steps: Optional[int]
    min_time: Optional[Tuple[float, DATETIME_UNITS]]

    _widgets = {
        "min_time": ReteTimeWidget(),
    }


@docstrings.get_sections(base="DropPaths")  # type: ignore
class DropPaths(DataProcessor):
    """
    Filter user paths based on the path length, removing the paths that are shorter than the
    specified number of events or time.

    Parameters
    ----------
    min_steps : int, optional
        Minimum user path length measured in the number of events.
    min_time : Tuple(float, :numpy_link:`DATETIME_UNITS<>`), optional
        Minimum user path length and its unit of measure.

    Returns
    -------
    Eventstream
        ``Eventstream`` with events that should be deleted from input ``eventstream`` marked ``_deleted=True``.

    Raises
    ------
    ValueError
        If both of ``min_steps`` and ``min_time`` are empty or both are given.

    See Also
    --------
    .TimedeltaHist : Plot the distribution of the time deltas between two events.
    .UserLifetimeHist : Plot the distribution of user lifetimes.

    Notes
    -----
    See :doc:`Data processors user guide</user_guides/dataprocessors>` for the details.
    """

    params: DropPathsParams

    @time_performance(
        scope="drop_paths",
        event_name="init",
    )
    def __init__(self, params: DropPathsParams):
        super().__init__(params=params)

    @time_performance(
        scope="drop_paths",
        event_name="apply",
    )
    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        user_col = schema.user_id
        time_col = schema.event_timestamp

        min_time, time_unit = None, None
        min_steps = self.params.min_steps

        if self.params.min_time:
            min_time, time_unit = self.params.min_time

        if min_steps and min_time:
            raise ValueError("min_steps and min_time parameters cannot be used simultaneously!")

        if not min_steps and not min_time:
            raise ValueError("Either min_steps or min_time must be specified!")

        if min_time and time_unit:
            userpath = df.groupby(user_col)[time_col].agg(start=np.min, end=np.max)  # type: ignore
            mask_ = (userpath["end"] - userpath["start"]) / np.timedelta64(1, time_unit) < min_time  # type: ignore

        else:
            userpath = df.groupby([user_col])[[time_col]].nunique().rename(columns={time_col: "length"})
            mask_ = userpath["length"] < min_steps

        users_to_delete = userpath[mask_].index
        result = df[~df[user_col].isin(users_to_delete)]

        collect_data_performance(
            scope="drop_paths",
            event_name="metadata",
            called_params=self.to_dict()["values"],
            not_hash_values=["min_time"],
            performance_data={
                "parent": {
                    "shape": df.shape,
                    "hash": hash_dataframe(df),
                },
                "child": {
                    "shape": result.shape,
                    "hash": hash_dataframe(result),
                },
            },
        )

        return result



import uuid
from typing import Optional

from jinja2 import Environment, FileSystemLoader, PackageLoader, Template


class PreprocessingGraphRenderer:
    __template: Template
    __environment: Environment

    def __init__(self) -> None:
        self.__environment = Environment(
            loader=PackageLoader(package_name="retentioneering", package_path="templates"),
        )
        self.__template = self.__environment.get_template("preprocessing_graph/template.html")

    def show(
        self,
        server_id: str,
        env: str,
        width: int,
        height: int,
        graph_url: str,
        graph_style_url: str,
        tracking_hardware_id: str,
        tracking_eventstream_index: int,
        tracking_scope: str,
        kernel_id: Optional[str] = None,
    ) -> str:
        return self.__template.render(
            server_id=server_id,
            env=env,
            block_id=str(uuid.uuid4()),
            width=width,
            height=height,
            graph_url=graph_url,
            graph_style_url=graph_style_url,
            tracking_hardware_id=tracking_hardware_id,
            kernel_id=kernel_id,
            tracking_eventstream_index=tracking_eventstream_index,
            tracking_scope=tracking_scope,
        )



from __future__ import annotations

from typing import Callable, Optional

from pandas import DataFrame, Series

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from retentioneering.utils.doc_substitution import docstrings

from ..types import EventstreamSchemaType, EventstreamType


class PipeHelperMixin:
    @docstrings.with_indent(12)
    @time_performance(  # type: ignore
        scope="pipe",
        event_name="helper",
        event_value="combine",
    )
    def pipe(
        self: EventstreamType, func: Callable[[DataFrame, Optional[EventstreamSchemaType]], DataFrame]
    ) -> EventstreamType:
        """
        Modify an input eventstream in an arbitrary way by applying given function.
        The function must accept a DataFrame associated with the input eventstream
        and return a new state of the modified eventstream.

        Parameters
        ----------
            %(Pipe.parameters)s

        Returns
        -------
        Eventstream
            Resulting eventstream
        """
        calling_params = {
            "func": func,
        }

        # avoid circular import
        from retentioneering.data_processors_lib import Pipe, PipeParams
        from retentioneering.preprocessing_graph import PreprocessingGraph
        from retentioneering.preprocessing_graph.nodes import EventsNode

        p = PreprocessingGraph(source_stream=self)  # type: ignore

        node = EventsNode(processor=Pipe(params=PipeParams(func=func)))  # type: ignore
        p.add_node(node=node, parents=[p.root])
        result = p.combine(node)
        del p
        collect_data_performance(
            scope="pipe",
            event_name="metadata",
            called_params=calling_params,
            performance_data={},
            eventstream_index=self._eventstream_index,
            parent_eventstream_index=self._eventstream_index,
            child_eventstream_index=result._eventstream_index,
        )

        return result



from .column import *
from .edges import *
from .env import *
from .nodes import *
from .normalization import *
from .recalculation_result import *
from .serialized_state import *
from .settings import *
from .state_changes import *
from .threshold import *
from .tracker import *
from .weight import *



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Literal, TypedDict, Union

from typing_extensions import NotRequired, TypeAlias

from .column import ColumnNumericMap

NodeId: TypeAlias = str

NodeTargetId: TypeAlias = Literal["suit_node", "positive", "negative", "source"]

NodeType: TypeAlias = str

NodesSortField = Literal["name", "weight", "custom"]

NodesSortOrder = Literal["asc", "desc", "custom"]


class NodeItem(TypedDict):
    id: NodeId
    name: str
    x: NotRequired[float]
    y: NotRequired[float]
    # TODO: It should be enum/literal
    visibilityRule: NotRequired[str]
    size: ColumnNumericMap
    weight: ColumnNumericMap
    targetId: NotRequired[NodeTargetId]
    isGroup: NotRequired[bool]
    isAggregated: NotRequired[bool]
    children: list[NodeId]
    parentNodeId: NotRequired[NodeId]
    description: NotRequired[str]
    customColor: NotRequired[str]


@dataclass
class RecalculationNodeItem:
    id: NodeId
    size: ColumnNumericMap
    weight: ColumnNumericMap
    targetId: NodeTargetId | None


@dataclass
class NodeAggregationData:
    ids: list[NodeId]
    weight: ColumnNumericMap


NodeAggregation: TypeAlias = Dict[NodeId, NodeAggregationData]

NodesCustomColors = Dict[NodeId, str]

NodeTargetEdgeDirection: TypeAlias = Literal["in", "out", "both"]

NodeTargetPosition: TypeAlias = Literal["top-right", "bottom-right", "top-left", "bottom-left"]


class NodeTarget(TypedDict):
    id: NodeTargetId
    name: str
    ignoreThreshold: NotRequired[bool]
    edgeDirection: NotRequired[NodeTargetEdgeDirection]
    position: NotRequired[NodeTargetPosition]
    description: NotRequired[str]


TargetToNodesMap: TypeAlias = Dict[NodeTargetId, Union[NodeId, List[NodeId]]]


class NodePoint(TypedDict):
    x: Union[int, float]
    y: Union[int, float]


NodesPoints: TypeAlias = Dict[NodeId, NodePoint]



from __future__ import annotations

from typing import List, Literal, Union

import pandas as pd

from retentioneering.backend.tracker import collect_data_performance, time_performance
from retentioneering.data_processor import DataProcessor
from retentioneering.eventstream.types import EventstreamSchemaType
from retentioneering.params_model import ParamsModel
from retentioneering.utils.doc_substitution import docstrings
from retentioneering.utils.hash_object import hash_dataframe
from retentioneering.widget.widgets import ListOfUsers


class LabelNewUsersParams(ParamsModel):
    """
    A class with parameters for :py:class:`.LabelNewUsers` class.
    """

    new_users_list: Union[List[int], List[str], Literal["all"]]
    _widgets = {"new_users_list": ListOfUsers()}


@docstrings.get_sections(
    base="LabelNewUsers",
)  # type: ignore
class LabelNewUsers(DataProcessor):
    """
    Create a new synthetic event for each user:
    ``new_user`` or ``existing_user``.

    Parameters
    ----------
    new_users_list : list of int or list of str or `all`
        If the `list of user_ids` is given - ``new_user`` event will be created for each user from the list.
        Event ``existing_user`` will be added to the rest of the users.

        If ``all`` - ``new_user`` synthetic event will be created for all users from the input ``eventstream``.

    Returns
    -------
    Eventstream
        Eventstream with new synthetic events, one for each user:

        +-----------------+-----------------+------------------------+
        | **event_name**  | **event_type**  | **timestamp**          |
        +-----------------+-----------------+------------------------+
        | new_user        | new_user        | first_event            |
        +-----------------+-----------------+------------------------+
        | existing_user   | existing_user   | first_event            |
        +-----------------+-----------------+------------------------+

    Notes
    -----
    See :doc:`Data processors user guide</user_guides/dataprocessors>` for the details.

    """

    params: LabelNewUsersParams

    @time_performance(
        scope="label_new_users",
        event_name="init",
    )
    def __init__(self, params: LabelNewUsersParams):
        super().__init__(params=params)

    @time_performance(  # type: ignore
        scope="label_new_users",
        event_name="apply",
    )
    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        user_col = schema.user_id
        type_col = schema.event_type
        event_col = schema.event_name
        new_users_list = self.params.new_users_list

        matched_events = df.groupby(user_col, as_index=False).first()

        if new_users_list == "all":
            matched_events[type_col] = "new_user"
            matched_events[event_col] = "new_user"
        else:
            new_user_mask = matched_events[user_col].isin(new_users_list)
            matched_events.loc[new_user_mask, type_col] = "new_user"  # type: ignore
            matched_events.loc[~new_user_mask, type_col] = "existing_user"  # type: ignore
            matched_events[event_col] = matched_events[type_col]

        result = pd.concat([df, matched_events])

        collect_data_performance(
            scope="label_new_users",
            event_name="metadata",
            called_params={"new_users_list": new_users_list},
            performance_data={
                "parent": {
                    "shape": df.shape,
                    "hash": hash_dataframe(df),
                },
                "child": {
                    "shape": result.shape,
                    "hash": hash_dataframe(result),
                },
            },
        )

        return result



from __future__ import annotations

import collections
import collections.abc
import typing
from abc import abstractmethod
from dataclasses import field
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Protocol,
    Tuple,
    TypedDict,
    Union,
)

import pandas as pd

from retentioneering.common.constants import FEATURE_TYPES, PATH_METRIC_TYPES

IndexOrder = List[Optional[str]]
OrderDict = Dict[str, int]


class EventstreamType(Protocol):
    schema: EventstreamSchemaType
    index_order: OrderDict
    __raw_data_schema: RawDataSchemaType
    __events: pd.DataFrame | pd.Series[Any]

    @abstractmethod
    def copy(self) -> EventstreamType: ...

    @property
    @abstractmethod
    def _eventstream_index(self) -> int: ...

    @abstractmethod
    def append_eventstream(self, eventstream: EventstreamType) -> None: ...

    @property
    @abstractmethod
    def _hash(self) -> str: ...

    @abstractmethod
    def to_dataframe(self, copy: bool = False, drop_segment_events: bool = True) -> pd.DataFrame: ...

    @abstractmethod
    def add_custom_col(self, name: str, data: pd.Series[Any] | None) -> None: ...

    @abstractmethod
    def segment_map(
        stream: EventstreamType,
        label: str,
        index: Literal["path_id", "segment_value"],
        resolve_collision: Optional[Literal["majority", "last"]] = None,
    ) -> pd.DataFrame | pd.Series: ...

    @abstractmethod
    def extract_features(
        self,
        feature_type: FEATURE_TYPES,
        ngram_range: Tuple[int, int],
        path_id_col: str | None = None,
        col_suffix: str | None = None,
    ) -> pd.DataFrame: ...

    @abstractmethod
    def path_metrics(
        self,
        metrics: Tuple[PATH_METRIC_TYPES, str] | List,
        path_id_col: str | None = None,
    ) -> pd.DataFrame | pd.Series: ...

    @abstractmethod
    def add_segment(self, segment: AddSegmentType, name: Optional[str] = None) -> EventstreamType: ...


class EventstreamSchemaType(Protocol):
    custom_cols: List[str] = field(default_factory=list)
    user_id: str = "user_id"
    event_timestamp: str = "event_timestamp"
    event_name: str = "event_name"
    event_index: str = "event_index"
    event_type: str = "event_type"
    event_id: str = "event_id"

    @abstractmethod
    def copy(self) -> EventstreamSchemaType: ...

    @abstractmethod
    def is_equal(self, schema: EventstreamSchemaType) -> bool: ...

    @abstractmethod
    def get_default_cols(self) -> List[str]: ...

    @abstractmethod
    def get_cols(self) -> list[str]: ...

    @abstractmethod
    def to_raw_data_schema(self, event_id: bool = False, event_index: bool = False) -> RawDataSchemaType: ...


class RawDataCustomColSchema(TypedDict):
    raw_data_col: str
    custom_col: str


class RawDataSchemaType(Protocol):
    event_name: str = "event"
    event_timestamp: str = "timestamp"
    user_id: str = "user_id"
    event_index: Optional[str] = None
    event_type: Optional[str] = None
    event_id: Optional[str] = None
    custom_cols: List[RawDataCustomColSchema] = field(default_factory=list)

    @abstractmethod
    def get_default_cols(self) -> List[str]: ...

    @abstractmethod
    def copy(self) -> RawDataSchemaType: ...


class SeriesWithValidator(pd.Series):  # pyright: ignore [reportUntypedBaseClass]
    @classmethod
    def __get_validators__(
        cls: typing.Type["SeriesWithValidator"],
    ) -> collections.abc.Iterable[collections.abc.Callable]:
        yield cls.validate_custom_class

    @classmethod
    def validate_custom_class(cls: typing.Type["SeriesWithValidator"], passed_value: typing.Any) -> pd.Series:
        if isinstance(passed_value, pd.Series):
            return passed_value

        raise ValueError


AddSegmentType = Union[str, Callable, SeriesWithValidator]
SplitExpr = Tuple[str, Any, Any]
SegmentFilterExpr = Tuple[str, Union[str, list]]
UserListType = List[Union[str, int]]
UserGroupsType = List[UserListType]
UserGroupsNamesType = Union[List[str], Tuple[str, str]]



from __future__ import annotations

import pandas as pd

from retentioneering.backend.tracker import collect_data_performance, time_performance
from retentioneering.data_processor import DataProcessor
from retentioneering.eventstream.segments import _get_segment_mask
from retentioneering.eventstream.types import EventstreamSchemaType, EventstreamType
from retentioneering.params_model import ParamsModel
from retentioneering.utils.doc_substitution import docstrings
from retentioneering.utils.hash_object import hash_dataframe


class DropSegmentParams(ParamsModel):
    name: str


@docstrings.get_sections(base="DropSegment")  # type: ignore
class DropSegment(DataProcessor):
    """
    Remove segment synthetic events from eventstream.

    Parameters
    ----------
    name : str
        Segment name to remove.

    Returns
    -------
    EventstreamType
        Eventstream with removed segment.
    """

    params: DropSegmentParams

    @time_performance(scope="drop_segment", event_name="init")
    def __init__(self, params: DropSegmentParams) -> None:
        super().__init__(params=params)

    @time_performance(scope="drop_segment", event_name="apply")
    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        hash_before = hash_dataframe(df)
        shape_before = df.shape
        mask = _get_segment_mask(df, schema, self.params.name)

        df.drop(df[mask].index, inplace=True)

        collect_data_performance(
            scope="drop_segment",
            event_name="metadata",
            called_params=self.to_dict()["values"],
            performance_data={
                "parent": {
                    "shape": shape_before,
                    "hash": hash_before,
                },
                "child": {
                    "shape": df.shape,
                    "hash": hash_dataframe(df),
                },
            },
        )

        return df



from __future__ import annotations

from typing import Any, Literal, Tuple

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from retentioneering.common.constants import DATETIME_UNITS, DATETIME_UNITS_LIST
from retentioneering.eventstream.types import EventstreamType

# @TODO Подумать над сокращением списка поддерживаемых типов для когорт? dpanina


class Cohorts:
    """
    A class that provides methods for cohort analysis. The users are split into groups
    depending on the time of their first appearance in the eventstream; thus each user is
    associated with some ``cohort_group``. Retention rates of the active users
    belonging to each ``cohort_group`` are  calculated within each ``cohort_period``.

    Parameters
    ----------
    eventstream : EventstreamType


    See Also
    --------
    .Eventstream.cohorts : Call Cohorts tool as an eventstream method.
    .EventTimestampHist : Plot the distribution of events over time.
    .UserLifetimeHist : Plot the distribution of user lifetimes.

    Notes
    -----
    See :doc:`Cohorts user guide</user_guides/cohorts>` for the details.

    """

    __eventstream: EventstreamType
    cohort_start_unit: DATETIME_UNITS
    cohort_period: int
    cohort_period_unit: DATETIME_UNITS

    average: bool
    cut_bottom: int
    cut_right: int
    cut_diagonal: int
    _cohort_matrix_result: pd.DataFrame

    @time_performance(
        scope="cohorts",
        event_name="init",
    )
    def __init__(self, eventstream: EventstreamType):
        self.__eventstream = eventstream
        self.user_col = self.__eventstream.schema.user_id
        self.event_col = self.__eventstream.schema.event_name
        self.time_col = self.__eventstream.schema.event_timestamp

        self._cohort_matrix_result = pd.DataFrame()

    def _add_min_date(
        self,
        data: pd.DataFrame,
        cohort_start_unit: DATETIME_UNITS,
        cohort_period: int,
        cohort_period_unit: DATETIME_UNITS,
    ) -> pd.DataFrame:
        freq = cohort_start_unit
        data["user_min_date_gr"] = data.groupby(self.user_col)[self.time_col].transform("min")
        min_cohort_date = data["user_min_date_gr"].min().to_period(freq).start_time
        max_cohort_date = data["user_min_date_gr"].max()
        if DATETIME_UNITS_LIST.index(cohort_start_unit) < DATETIME_UNITS_LIST.index(cohort_period_unit):
            freq = cohort_period_unit

        if freq == "W":
            freq = "D"

        data["user_min_date_gr"] = data["user_min_date_gr"].dt.to_period(freq)

        step = np.timedelta64(cohort_period, cohort_period_unit)
        start_point = np.datetime64(min_cohort_date, freq)
        end_point = np.datetime64(max_cohort_date, freq) + np.timedelta64(cohort_period, cohort_period_unit)

        coh_groups_start_dates = np.arange(start_point, end_point, step)
        coh_groups_start_dates = pd.to_datetime(coh_groups_start_dates).to_period(freq)
        if max_cohort_date < coh_groups_start_dates[-1].start_time:  # type: ignore
            coh_groups_start_dates = coh_groups_start_dates[:-1]  # type: ignore

        cohorts_list = pd.DataFrame(
            data=coh_groups_start_dates, index=None, columns=["CohortGroup"]  # type: ignore
        ).reset_index()
        cohorts_list.columns = ["CohortGroupNum", "CohortGroup"]  # type: ignore
        cohorts_list["CohortGroupNum"] += 1

        data["OrderPeriod"] = data[self.time_col].dt.to_period(freq)
        start_int = pd.Series(min_cohort_date.to_period(freq=freq)).astype(int)[0]

        converter_freq = np.timedelta64(cohort_period, cohort_period_unit)
        converter_freq_ = converter_freq.astype(f"timedelta64[{freq}]").astype(int)
        data["CohortGroupNum"] = (data["user_min_date_gr"].astype(int) - start_int + converter_freq_) // converter_freq_

        data = data.merge(cohorts_list, on="CohortGroupNum", how="left")

        data["CohortPeriod"] = (
            (data["OrderPeriod"].astype(int) - (data["CohortGroup"].astype(int) + converter_freq_))  # type: ignore
            // converter_freq_
        ) + 1

        return data

    @staticmethod
    def _cut_cohort_matrix(
        df: pd.DataFrame, cut_bottom: int = 0, cut_right: int = 0, cut_diagonal: int = 0
    ) -> pd.DataFrame:
        for row in df.index:
            df.loc[row, max(0, df.loc[row].notna()[::-1].idxmax() + 1 - cut_diagonal) :] = None  # type: ignore

        return df.iloc[: len(df) - cut_bottom, : len(df.columns) - cut_right]

    @time_performance(
        scope="cohorts",
        event_name="fit",
    )
    def fit(
        self,
        cohort_start_unit: DATETIME_UNITS,
        cohort_period: Tuple[int, DATETIME_UNITS],
        average: bool = True,
        cut_bottom: int = 0,
        cut_right: int = 0,
        cut_diagonal: int = 0,
    ) -> None:
        """
        Calculates the cohort internal values with the defined parameters.
        Applying ``fit`` method is necessary for the following usage
        of any visualization or descriptive ``Cohorts`` methods.

        Parameters
        ----------
        cohort_start_unit : :numpy_link:`DATETIME_UNITS<>`
            The way of rounding and formatting of the moment from which the cohort count begins.
            The minimum timestamp is rounded down to the selected datetime unit.

            For example:
            assume we have an eventstream with the following minimum timestamp - "2021-12-28 09:08:34.432456".
            The result of roundings with different ``DATETIME_UNITS`` is shown in the table below:

            +------------------------+-------------------------+
            | **cohort_start_unit**  | **cohort_start_moment** |
            +------------------------+-------------------------+
            | Y                      |  2021-01-01 00:00:00    |
            +------------------------+-------------------------+
            | M                      |  2021-12-01 00:00:00    |
            +------------------------+-------------------------+
            | W                      |  2021-12-27 00:00:00    |
            +------------------------+-------------------------+
            | D                      |  2021-08-28 00:00:00    |
            +------------------------+-------------------------+

        cohort_period : Tuple(int, :numpy_link:`DATETIME_UNITS<>`)
            The cohort_period size and its ``DATETIME_UNIT``. This parameter is used in calculating:

            - Start moments for each cohort from the moment specified with the ``cohort_start_unit`` parameter
            - Cohort periods for each cohort from its start moment.
        average : bool, default True
            - If ``True`` - calculating average for each cohort period.
            - If ``False`` - averaged values aren't calculated.
        cut_bottom : int, default 0
            Drop 'n' rows from the bottom of the cohort matrix.
            Average is recalculated.
        cut_right : int, default 0
            Drop 'n' columns from the right side of the cohort matrix.
            Average is recalculated.
        cut_diagonal : int, default 0
            Replace values in 'n' diagonals (last period-group cells) with ``np.nan``.
            Average is recalculated.

        Notes
        -----
        Parameters ``cohort_start_unit`` and ``cohort_period`` should be consistent.
        Due to "Y" and "M" being non-fixed types, it can be used only with each other
        or if ``cohort_period_unit`` is more detailed than ``cohort_start_unit``.
        More information - :numpy_timedelta_link:`about numpy timedelta<>`

        Only cohorts with at least 1 user in some period are shown.

        See :doc:`Cohorts user guide</user_guides/cohorts>` for the details.

        """
        called_params = {
            "cohort_start_unit": cohort_start_unit,
            "cohort_period": cohort_period,
            "average": average,
            "cut_bottom": cut_bottom,
            "cut_right": cut_right,
            "cut_diagonal": cut_diagonal,
        }
        not_hash_values = ["cohort_start_unit", "cohort_period"]

        data = self.__eventstream.to_dataframe()
        self.average = average
        self.cohort_start_unit = cohort_start_unit
        self.cohort_period, self.cohort_period_unit = cohort_period
        self.cut_bottom = cut_bottom
        self.cut_right = cut_right
        self.cut_diagonal = cut_diagonal

        if self.cohort_period <= 0:
            raise ValueError("cohort_period should be positive integer!")

        # @TODO добавить ссылку на numpy с объяснением. dpanina
        if self.cohort_period_unit in ["Y", "M"] and self.cohort_start_unit not in ["Y", "M"]:
            raise ValueError(
                """Parameters ``cohort_start_unit`` and ``cohort_period`` should be consistent.
                                 Due to "Y" and "M" are non-fixed types it can be used only with each other
                                 or if ``cohort_period_unit`` is more detailed than ``cohort_start_unit``!"""
            )

        df = self._add_min_date(
            data=data,
            cohort_start_unit=self.cohort_start_unit,
            cohort_period=self.cohort_period,
            cohort_period_unit=self.cohort_period_unit,
        )

        cohorts = df.groupby(["CohortGroup", "CohortPeriod"])[[self.user_col]].nunique()
        cohorts.reset_index(inplace=True)

        cohorts.rename(columns={self.user_col: "TotalUsers"}, inplace=True)
        cohorts.set_index(["CohortGroup", "CohortPeriod"], inplace=True)
        cohort_group_size = cohorts["TotalUsers"].groupby(level=0).first()
        cohorts.reset_index(inplace=True)
        user_retention = (
            cohorts.pivot(index="CohortPeriod", columns="CohortGroup", values="TotalUsers").divide(
                cohort_group_size, axis=1
            )
        ).T

        user_retention = self._cut_cohort_matrix(
            df=user_retention, cut_diagonal=self.cut_diagonal, cut_bottom=self.cut_bottom, cut_right=self.cut_right
        )
        user_retention.index = user_retention.index.astype(str)
        if self.average:
            user_retention.loc["Average"] = user_retention.mean()

        self._cohort_matrix_result = user_retention
        collect_data_performance(
            scope="cohorts",
            event_name="metadata",
            called_params=called_params,
            not_hash_values=not_hash_values,
            performance_data={"shape": self._cohort_matrix_result.shape},
            eventstream_index=self.__eventstream._eventstream_index,
        )

    @time_performance(
        scope="cohorts",
        event_name="heatmap",
    )
    def heatmap(self, width: float = 5.0, height: float = 5.0) -> Any:
        """
        Builds a heatmap based on the calculated cohort matrix values.
        Should be used after :py:func:`fit`.

        Parameters
        ----------
        width : float, default 5.0
            Width of the figure in inches.
        height : float, default 5.0
            Height of the figure in inches.


        Returns
        -------
        matplotlib.axes.Axes
        """
        called_params = {
            "width": width,
            "height": height,
        }

        df = self._cohort_matrix_result
        figsize = (width, height)
        figure, ax = plt.subplots(figsize=figsize)
        sns.heatmap(df, annot=True, fmt=".1%", linewidths=1, linecolor="gray", ax=ax)
        collect_data_performance(
            scope="cohorts",
            event_name="metadata",
            called_params=called_params,
            performance_data={"shape": self._cohort_matrix_result.shape},
            eventstream_index=self.__eventstream._eventstream_index,
        )

        return ax

    @time_performance(
        scope="cohorts",
        event_name="lineplot",
    )
    def lineplot(
        self, plot_type: Literal["cohorts", "average", "all"] = "cohorts", width: float = 7.0, height: float = 5.0
    ) -> Any:
        """
        Create a chart representing each cohort dynamics over time.
        Should be used after :py:func:`fit`.

        Parameters
        ----------
        plot_type: 'cohorts', 'average' or 'all'
            - if ``cohorts`` - shows a lineplot for each cohort,
            - if ``average`` - shows a lineplot only for the average values over all the cohorts,
            - if ``all`` - shows a lineplot for each cohort and also for their average values.
        width : float, default 7.0
            Width of the figure in inches.
        height : float, default 5.0
            Height of the figure in inches.

        Returns
        -------
        matplotlib.axes.Axes

        """
        if plot_type not in ["cohorts", "average", "all"]:
            raise ValueError("plot_type parameter should be 'cohorts', 'average' or 'all'!")
        called_params = {
            "plot_type": plot_type,
            "width": width,
            "height": height,
        }
        not_hash_values = ["plot_type"]

        figsize = (width, height)
        df_matrix = self._cohort_matrix_result
        df_wo_average = df_matrix[df_matrix.index != "Average"]  # type: ignore
        if plot_type in ["all", "average"] and "Average" not in df_matrix.index:  # type: ignore
            df_matrix.loc["Average"] = df_matrix.mean()  # type: ignore
        df_average = df_matrix[df_matrix.index == "Average"]  # type: ignore
        figure, ax = plt.subplots(figsize=figsize)
        if plot_type == "all":
            sns.lineplot(df_wo_average.T, lw=1.5, ax=ax)
            sns.lineplot(df_average.T, lw=2.5, palette=["red"], marker="X", markersize=8, alpha=0.6, ax=ax)

        if plot_type == "average":
            sns.lineplot(df_average.T, lw=2.5, palette=["red"], marker="X", markersize=8, alpha=0.6, ax=ax)

        if plot_type == "cohorts":
            sns.lineplot(df_wo_average.T, lw=1.5, ax=ax)

        ax.legend(loc="upper left", bbox_to_anchor=(1, 1))
        ax.set_xlabel("Period from the start of observation")
        ax.set_ylabel("Share of active users")

        collect_data_performance(
            scope="cohorts",
            event_name="metadata",
            called_params=called_params,
            not_hash_values=not_hash_values,
            performance_data={"shape": self._cohort_matrix_result.shape},
            eventstream_index=self.__eventstream._eventstream_index,
        )
        return ax

    @property
    @time_performance(
        scope="cohorts",
        event_name="values",
    )
    def values(self) -> pd.DataFrame:
        """
        Returns a pd.DataFrame representing the calculated cohort matrix values.
        Should be used after :py:func:`fit`.

        Returns
        -------
        pd.DataFrame

        """
        return self._cohort_matrix_result

    @property
    @time_performance(  # type: ignore
        scope="cohorts",
        event_name="params",
    )
    def params(self) -> dict[str, DATETIME_UNITS | tuple | bool | int | None]:
        """
        Returns the parameters used for the last fitting.
        Should be used after :py:func:`fit`.

        """
        return {
            "cohort_start_unit": self.cohort_start_unit,
            "cohort_period": (self.cohort_period, self.cohort_period_unit),
            "average": self.average,
            "cut_bottom": self.cut_bottom,
            "cut_right": self.cut_right,
            "cut_diagonal": self.cut_diagonal,
        }



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Any, List, Optional, cast

import networkx
import pandas as pd
from IPython.core.display import HTML, DisplayHandle, display
from pydantic import ValidationError

from retentioneering import RETE_CONFIG
from retentioneering.backend import JupyterServer, ServerManager
from retentioneering.backend.callback import list_dataprocessor, list_dataprocessor_mock
from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
    tracker,
)
from retentioneering.eventstream.types import EventstreamType
from retentioneering.exceptions.server import ServerErrorWithResponse
from retentioneering.exceptions.widget import WidgetParseError
from retentioneering.preprocessing_graph.interface import (
    CombineHandlerPayload,
    CreateNodeErrorDesc,
    FieldErrorDesc,
    NodeLink,
    Payload,
    PreprocessingGraphConfig,
)
from retentioneering.preprocessing_graph.nodes import (
    EventsNode,
    MergeNode,
    Node,
    SourceNode,
    build_node,
)
from retentioneering.templates import PreprocessingGraphRenderer


class PreprocessingGraph:
    """
    Collection of methods for preprocessing graph construction and calculation.

    Parameters
    ----------
    source_stream : EventstreamType
        Source eventstream.

    Notes
    -----
    See :doc:`Preprocessing user guide</user_guides/preprocessing>` for the details.

    """

    DEFAULT_GRAPH_URL = "https://static.server.retentioneering.com/package/@rete/preprocessing-graph/version/2/dist/preprocessing-graph.umd.js"
    DEFAULT_GRAPH_STYLE_URL = (
        "https://static.server.retentioneering.com/package/@rete/preprocessing-graph/version/2/dist/style.css"
    )
    root: SourceNode
    _combine_result: EventstreamType | None

    _ngraph: networkx.DiGraph
    __server_manager: ServerManager | None = None
    __server: JupyterServer | None = None

    @time_performance(  # type: ignore
        scope="preprocessing_graph",
        event_name="init",
    )
    def __init__(self, source_stream: EventstreamType) -> None:
        self._source_stream = source_stream
        self.root = SourceNode(source=source_stream)
        self._combine_result = None
        self._ngraph = networkx.DiGraph()
        self._ngraph.add_node(self.root)

    def add_node(self, node: Node, parents: List[Node]) -> None:
        """
        Add node to ``PreprocessingGraph`` instance.

        Parameters
        ----------
        node : Node
            An instance of either ``EventsNode`` or ``MergeNode``.
        parents : list of Nodes

            - If ``node`` is ``EventsNode`` - only 1 parent must be defined.
            - If ``node`` is ``MergeNode`` - at least 2 parents have to be defined.

        Returns
        -------
        None

        See Also
        --------
        PreprocessingGraph.combine : Start PreprocessingGraph recalculation.
        .EventsNode : Regular nodes of a preprocessing graph.
        .MergeNode : Merge nodes of a preprocessing graph.

        """
        self.__valiate_already_exists(node)
        self.__validate_not_found(parents)

        if node.events is not None:
            self.__validate_schema(node.events)

        if not isinstance(node, MergeNode) and len(parents) > 1:
            raise ValueError("multiple parents are only allowed for merge nodes!")

        self._ngraph.add_node(node)

        for parent in parents:
            self._ngraph.add_edge(parent, node)

    @time_performance(
        scope="preprocessing_graph",
        event_name="combine",
    )
    def combine(self, node: Node) -> EventstreamType:
        """
        Run calculations from the ``SourceNode`` up to the specified ``node``.

        Parameters
        ----------
        node : Node
            Instance of either ``SourceNode``, ``EventsNode`` or ``MergeNode``.

        Returns
        -------
        EventstreamType
            ``Eventstream`` with all changes applied by data processors.
        """
        self.__validate_not_found([node])

        if isinstance(node, SourceNode):
            result_eventstream = node.events.copy()

        elif isinstance(node, EventsNode):
            result_eventstream = self._combine_events_node(node)

        else:
            result_eventstream = self._combine_merge_node(node)

        collect_data_performance(
            scope="preprocessing_graph",
            called_params={"node": node},
            performance_data={
                "parent": {
                    "index": getattr(self.__get_node_eventstream(node), "_eventstream_index", None),
                    "hash": getattr(self.__get_node_eventstream(node), "_hash", None),
                },
                "child": {"index": result_eventstream._eventstream_index, "hash": result_eventstream._hash},
            },
            eventstream_index=getattr(self.__get_node_eventstream(node), "_eventstream_index", None),
            parent_eventstream_index=getattr(self.__get_node_eventstream(node), "_eventstream_index", None),
            child_eventstream_index=result_eventstream._eventstream_index,
        )
        return result_eventstream

    def __get_node_eventstream(self, node: Node) -> pd.DataFrame | pd.Series | None:
        if events := getattr(node, "events", None):
            return events
        return None

    def _combine_events_node(self, node: EventsNode) -> EventstreamType:
        parent = self._get_events_node_parent(node)
        parent_events = self.combine(parent)
        return node.processor._get_new_data(eventstream=parent_events)

    def _combine_merge_node(self, node: MergeNode) -> EventstreamType:
        parents = self._get_merge_node_parents(node)
        curr_eventstream: Optional[EventstreamType] = None

        for parent_node in parents:
            if curr_eventstream is None:
                curr_eventstream = self.combine(parent_node)
            else:
                new_eventstream = self.combine(parent_node)
                curr_eventstream.index_order = new_eventstream.index_order
                curr_eventstream.append_eventstream(new_eventstream)
        node.events = curr_eventstream

        return cast(EventstreamType, curr_eventstream)

    def get_parents(self, node: Node) -> List[Node]:
        """
        Show parents of the specified ``node``.

        Parameters
        ----------
        node : Node
            Instance of one of the classes SourceNode, EventsNode or MergeNode.

        Returns
        -------
        list of Nodes

        """
        self.__validate_not_found([node])
        parents: List[Node] = []

        for parent in self._ngraph.predecessors(node):
            parents.append(parent)
        return parents

    def _get_merge_node_parents(self, node: MergeNode) -> List[Node]:
        parents = self.get_parents(node)
        if len(parents) == 0:
            raise ValueError("orphan merge node!")

        return parents

    def _get_events_node_parent(self, node: EventsNode) -> Node:
        parents = self.get_parents(node)
        if len(parents) > 1:
            raise ValueError("invalid preprocessing graph: events node has more than 1 parent")

        return parents[0]

    def __validate_schema(self, eventstream: EventstreamType) -> bool:
        return self.root.events.schema.is_equal(eventstream.schema)

    def __valiate_already_exists(self, node: Node) -> None:
        if node in self._ngraph.nodes:
            raise ValueError("node already exists!")

    def __validate_not_found(self, nodes: List[Node]) -> None:
        for node in nodes:
            if node not in self._ngraph.nodes:
                raise ValueError("node not found!")

    @property
    def graph_url(self) -> str:
        env_url = os.getenv("RETE_PREPROCESSING_GRAPH_URL", "")
        return env_url if env_url else self.DEFAULT_GRAPH_URL

    @property
    def graph_style_url(self) -> str:
        env_style = os.getenv("RETE_PREPROCESSING_GRAPH_STYLE_URL", "")
        return env_style if env_style else self.DEFAULT_GRAPH_STYLE_URL

    @time_performance(
        scope="preprocessing_graph",
        event_name="display",
    )
    def display(self, width: int = 960, height: int = 600) -> DisplayHandle:
        """
        Show constructed ``PreprocessingGraph``.

        Parameters
        ----------
        width : int, default 960
            Width of plot in pixels.
        height : int, default 600
            Height of plot in pixels.

        Returns
        -------
            Rendered preprocessing graph.
        """
        if not self.__server_manager:
            self.__server_manager = ServerManager()

        if not self.__server:
            self.__server = self.__server_manager.create_server()
            self.__server.register_action("list-dataprocessor-mock", list_dataprocessor_mock)
            self.__server.register_action("list-dataprocessor", list_dataprocessor)
            self.__server.register_action("set-graph", self._set_graph_handler)
            self.__server.register_action("get-graph", self._export_handler)
            self.__server.register_action("combine", self._combine_handler)

        render = PreprocessingGraphRenderer()

        collect_data_performance(
            scope="preprocessing_graph",
            event_name="metadata",
            called_params={"width": width, "height": height},
            eventstream_index=self.root.events._eventstream_index,
        )

        return display(
            HTML(
                render.show(
                    server_id=self.__server.pk,
                    env="jupyter",
                    width=width,
                    height=height,
                    graph_url=self.graph_url,
                    graph_style_url=self.graph_style_url,
                    kernel_id="migrate_to_new_bridge",
                    tracking_hardware_id=RETE_CONFIG.user.pk,
                    tracking_eventstream_index=self._source_stream._eventstream_index,
                    tracking_scope="preprocessing_graph",
                )
            )
        )

    def _export_handler(self, payload: dict[str, Any]) -> dict:
        """
        Show ``PreprocessingGraph`` as a dict.

        Parameters
        ----------
        payload : dict

        Returns
        -------
        dict

        """
        source, target, link = "source", "target", "links"
        graph = self._ngraph
        data = {
            "directed": graph.is_directed(),
            "nodes": [n.export() for n in graph],
            link: [{source: u.pk, target: v.pk} for u, v, d in graph.edges(data=True)],
        }
        return data

    def __resolve_path(self, relative_path: str) -> str:
        path = Path(relative_path)
        absolute_path = str(path.resolve())
        return absolute_path

    @time_performance(
        scope="preprocessing_graph",
        event_name="export_to_file",
    )
    def export_to_file(self, filename: str) -> None:
        """
        Export constructed ``PreprocessingGraph`` to a json file.

        Parameters
        ----------
        filename : str
            Path to the json file.

        Returns
        -------
        None
        """

        # validate access rights to file

        graph_path = self.__resolve_path(filename)

        if not os.access(os.path.dirname(graph_path), os.W_OK):
            raise PermissionError(f"{os.path.dirname(graph_path)} is not writable")

        # export preprocessing_graph
        data = self._export_handler({})
        with open(graph_path, "w") as f:
            json.dump(data, f)

    @time_performance(
        scope="preprocessing_graph",
        event_name="import_from_file",
    )
    def import_from_file(self, filename: str) -> None:
        """
        Import constructed ``PreprocessingGraph`` from a json file.

        Parameters
        ----------
        filename : str
            Path to the json file.

        Returns
        -------
        None
        """
        graph_path = self.__resolve_path(filename)

        # validate access rights to file
        if not os.access(os.path.dirname(graph_path), os.R_OK):
            raise PermissionError(f"{os.path.dirname(graph_path)} is not readable")

        with open(graph_path, "r") as f:
            data = json.load(f)
        if validation_error := self._validate_payload(payload=data):
            raise ValueError("Invalid json file: %s" % validation_error)

        payload: Payload = Payload(**data)  # type: ignore
        self._set_graph_handler(payload=payload)

    def _combine_handler(self, payload: CombineHandlerPayload) -> None:
        node = self._find_node(payload["node_pk"])
        if not node:
            raise ServerErrorWithResponse(message="node not found!", type="unexpected_error")
        self._combine_result = self.combine(node)

    @property
    @time_performance(
        scope="preprocessing_graph",
        event_name="combine_result",
    )
    def combine_result(self) -> Optional[EventstreamType]:
        """
        Keep and get the last combining result from preprocessing graph GUI.

        Returns
        -------
        EventstreamType or None
            Preprocessed eventstream.

        """
        return self._combine_result

    def _set_graph_handler(self, payload: Payload) -> dict:
        current_graph = self._ngraph
        current_root = self.root

        def restore_graph() -> None:
            self._ngraph = current_graph
            self.root = current_root

        try:
            self._set_graph(payload=payload)
            return self._export_handler({})
        except ServerErrorWithResponse as err:
            restore_graph()
            raise err
        except Exception as err:
            restore_graph()
            raise ServerErrorWithResponse(message=str(err), type="unexpected_error")

    def _validate_payload(self, payload: Payload) -> None | Exception:
        try:
            config = PreprocessingGraphConfig(**payload)
            return None
        except Exception as err:
            return err

    def _set_graph(self, payload: Payload) -> None:
        """
        Payload example:

        {
            "nodes": [
                {
                    "name": "SourceNode",
                    "pk": "0dc3b706-e6cc-401e-96f7-6a45d3947d5c"
                },
                {
                    "name": "EventsNode",
                    "pk": "07921cb0-60b8-45af-928d-272d1b622b25",
                    "processor": {
                        "name": "SimpleGroup",
                        "values": {"event_name": "add_to_cart", "event_type": "group_alias"},
                    },
                },
                {
                    "name": "EventsNode",
                    "pk": "114251ae-0f03-45e6-a163-af51bb02dfd5",
                    "processor": {
                        "name": "SimpleGroup",
                        "values": {"event_name": "logout", "event_type": "group_alias"},
                    },
                },
            ],
            "links": [
                {
                    'source': '0dc3b706-e6cc-401e-96f7-6a45d3947d5c',
                    'target': '07921cb0-60b8-45af-928d-272d1b622b25'
                },
                {
                    'source': '07921cb0-60b8-45af-928d-272d1b622b25',
                    'target': '114251ae-0f03-45e6-a163-af51bb02dfd5'
                }
            ]
        }

        """
        errors: List[CreateNodeErrorDesc] = []
        nodes: List[Node] = []

        # create nodes & validate params
        for node in payload["nodes"]:
            node_pk = node["pk"]
            processor = node.get("processor", {})
            processor_name = processor.get("name", None) if processor else None
            processor_params = processor.get("values", None) if processor else None
            description = node.get("description", None)

            try:
                actual_node = build_node(
                    source_stream=self.root.events,
                    pk=node_pk,
                    node_name=node["name"],
                    processor_name=processor_name,
                    processor_params=processor_params,
                    descriptionn=description,
                )
                nodes.append(actual_node)
            except Exception as error:
                error_desc = self._build_node_error_desc(node_pk=node_pk, error=error)
                errors.append(error_desc)

        if errors:
            raise ServerErrorWithResponse(
                message="set preprocessing graph error", type="create_nodes_error", errors=errors
            )

        self._ngraph = networkx.DiGraph()

        # add nodes
        for created_node in nodes:
            if isinstance(created_node, SourceNode):
                self.root = created_node
            self._ngraph.add_node(created_node)

        # add links
        # @TODO: validate links (graph structure)
        for link in payload["links"]:
            source = self._find_node(link["source"])
            target = self._find_node(link["target"])
            if not source:
                raise ServerErrorWithResponse(message="source not found", type="create_link_error")
            if not target:
                raise ServerErrorWithResponse(message="target not found", type="create_link_error")
            self._ngraph.add_edge(source, target)

    def _build_node_error_desc(self, node_pk: str, error: Exception) -> CreateNodeErrorDesc:
        if isinstance(error, ValidationError):
            return self._build_pydantic_error_desc(
                node_pk=node_pk,
                validation_error_exception=error,
            )

        if isinstance(error, WidgetParseError):
            field_errors: List[FieldErrorDesc] = (
                [{"field": error.field_name, "msg": str(error)}] if error.field_name else []
            )
            return {
                "type": "node_error",
                "msg": str(error),
                "node_pk": node_pk,
                "fields_errors": field_errors,
            }

        return {
            "type": "node_error",
            "msg": str(error),
            "node_pk": node_pk,
            "fields_errors": [],
        }

    def _build_pydantic_error_desc(
        self, node_pk: str, validation_error_exception: ValidationError
    ) -> CreateNodeErrorDesc:
        raw_errs = validation_error_exception.errors()
        result_errors: List[FieldErrorDesc] = []

        for raw_err in raw_errs:
            loc = raw_err.get("loc", ())
            field = next(iter(loc), None)
            msg = raw_err.get("msg")

            result_errors.append(
                {
                    "field": str(field),
                    "msg": msg,
                }
            )

        return {"type": "node_error", "node_pk": node_pk, "msg": "node error", "fields_errors": result_errors}

    def _find_parents_by_links(self, target_node: str, link_list: list[NodeLink]) -> list[Node]:
        parents: list[str] = []
        for node in link_list:
            if node["target"] == target_node:
                parents.append(node["source"])

        parent_nodes = [self._find_node(parent) for parent in parents]
        return parent_nodes  # type: ignore

    def _find_node(self, pk: str) -> Node | None:
        for node in self._ngraph:
            if node.pk == pk:
                return node
        else:
            return None



from __future__ import annotations

import typing
from typing import Type

from retentioneering.utils.registry import RegistryValidationError, ReteRegistry

if typing.TYPE_CHECKING:
    from retentioneering.params_model import ParamsModel


class ParamsModelRegistry:
    REGISTRY: dict[str, "Type[ParamsModel]"] = {}  # type: ignore

    objects = "ParamsModel"

    def __setitem__(self, key: str, value: "Type[ParamsModel]") -> None:
        if key not in self.REGISTRY:
            self.REGISTRY[key] = value

    def __delitem__(self, key: str) -> None:
        del self.REGISTRY[key]

    @classmethod
    def get_registry(cls: typing.Type[ParamsModelRegistry]) -> dict:
        return cls.REGISTRY


params_model_registry = ParamsModelRegistry()


def register_params_model(cls: "Type[ParamsModel]") -> None:
    params_model_registry[cls.__name__] = cls


def unregister_params_model(cls: "Type[ParamsModel]") -> None:
    del params_model_registry[cls.__name__]



from __future__ import annotations

from typing import Any

import numpy as np
import pandas as pd

from retentioneering.backend.tracker import collect_data_performance, time_performance
from retentioneering.data_processor import DataProcessor
from retentioneering.eventstream.segments import (
    SEGMENT_DELIMITER,
    SEGMENT_TYPE,
    _create_segment_event,
    _extract_segment_values,
    _get_segment_mask,
)
from retentioneering.eventstream.types import (
    AddSegmentType,
    EventstreamSchemaType,
    EventstreamType,
)
from retentioneering.params_model import ParamsModel
from retentioneering.utils.doc_substitution import docstrings
from retentioneering.utils.hash_object import hash_dataframe
from retentioneering.widget.widgets import ListOfString, ReteFunction

ADD_SEGMENT_TIMEDELTA = pd.Timedelta(1, "microseconds")


class RemapSegmentParams(ParamsModel):
    name: str
    mapping: dict


@docstrings.get_sections(base="RemapSegment")  # type: ignore
class RemapSegment(DataProcessor):
    """
    Remap segment values for synthetic eventstream events.

    Parameters
    ----------
    name : str
        A segment name to remap.
    mapping : str
        A pandas.Series containing mapping rules. The series index relates to old segment values,
         and the values relate to new segment values.

    Returns
    -------
    EventstreamType
        Eventstream with remapped segment.
    """

    params: RemapSegmentParams

    @time_performance(scope="remap_segment", event_name="init")
    def __init__(self, params: RemapSegmentParams) -> None:
        super().__init__(params=params)

    @time_performance(scope="remap_segment", event_name="apply")
    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        def mapper(value: Any) -> Any:
            return value if not mapping.get(value) else mapping.get(value)

        name = self.params.name
        mapping = self.params.mapping
        hash_before = hash_dataframe(df)
        shape_before = df.shape

        mask = _get_segment_mask(df, schema, name)
        values = _extract_segment_values(df.loc[mask, schema.event_name])

        df.loc[mask, schema.event_name] = _create_segment_event(values.map(mapper), name)

        collect_data_performance(
            scope="remap_segment",
            event_name="metadata",
            called_params=self.to_dict()["values"],
            performance_data={
                "parent": {
                    "shape": shape_before,
                    "hash": hash_before,
                },
                "child": {
                    "shape": df.shape,
                    "hash": hash_dataframe(df),
                },
            },
        )
        return df



from __future__ import annotations

from ..types import EventstreamType


class RenameHelperMixin:
    def rename(self: EventstreamType, rules: list[dict[str, list[str] | str]]) -> EventstreamType:
        # avoid circular import
        from retentioneering.data_processors_lib import RenameParams, RenameProcessor
        from retentioneering.preprocessing_graph import PreprocessingGraph
        from retentioneering.preprocessing_graph.nodes import EventsNode

        p = PreprocessingGraph(source_stream=self)  # type: ignore

        node = EventsNode(processor=RenameProcessor(params=RenameParams(rules=rules)))  # type: ignore
        p.add_node(node=node, parents=[p.root])
        result = p.combine(node)
        del p
        return result



from __future__ import annotations

from typing import Literal, Optional

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from retentioneering.utils.doc_substitution import docstrings

from ..types import EventstreamType


class TruncatePathsHelperMixin:
    @docstrings.with_indent(12)
    @time_performance(  # type: ignore
        scope="truncate_paths",
        event_name="helper",
        event_value="combine",
    )
    def truncate_paths(
        self: EventstreamType,
        drop_before: Optional[str] = None,
        drop_after: Optional[str] = None,
        occurrence_before: Literal["first", "last"] = "first",
        occurrence_after: Literal["first", "last"] = "first",
        shift_before: int = 0,
        shift_after: int = 0,
        ignore_before: bool = False,
        ignore_after: bool = False,
        keep_synthetic: bool = False,
    ) -> EventstreamType:
        """
        A method of ``Eventstream`` class that truncates each user's path based on the
        specified event(s) and selected parameters.


        Parameters
        ----------
            %(TruncatePath.parameters)s

        Returns
        -------
        Eventstream
             Input ``eventstream`` with truncated paths.


        """
        # avoid circular import
        from retentioneering.data_processors_lib import (
            TruncatePaths,
            TruncatePathsParams,
        )
        from retentioneering.preprocessing_graph import PreprocessingGraph
        from retentioneering.preprocessing_graph.nodes import EventsNode

        p = PreprocessingGraph(source_stream=self)  # type: ignore
        params = {
            "drop_before": drop_before,
            "drop_after": drop_after,
            "occurrence_before": occurrence_before,
            "occurrence_after": occurrence_after,
            "shift_before": shift_before,
            "shift_after": shift_after,
            "ignore_before": ignore_before,
            "ignore_after": ignore_after,
            "keep_synthetic": keep_synthetic,
        }
        not_hash_values = ["occurrence_before", "occurrence_after"]

        node = EventsNode(processor=TruncatePaths(params=TruncatePathsParams(**params)))  # type: ignore
        p.add_node(node=node, parents=[p.root])
        result = p.combine(node)
        del p
        collect_data_performance(
            scope="truncate_paths",
            event_name="metadata",
            called_params=params,
            not_hash_values=not_hash_values,
            performance_data={},
            eventstream_index=self._eventstream_index,
            parent_eventstream_index=self._eventstream_index,
            child_eventstream_index=result._eventstream_index,
        )

        return result



from __future__ import annotations

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from retentioneering.utils.doc_substitution import docstrings

from ..types import EventstreamType


class AddStartEndEventsHelperMixin:
    @docstrings.dedent
    @time_performance(
        scope="add_start_end_events",
        event_name="helper",
        event_value="combine",
    )
    def add_start_end_events(self: EventstreamType) -> EventstreamType:
        """
        A method of ``Eventstream`` class that creates
        two synthetic events in each user's path: ``path_start`` and ``path_end``.

        Returns
        -------
        Eventstream
            Input ``eventstream`` with added synthetic events. See details :py:class:`.AddStartEndEvents`.


        """
        # avoid circular import
        from retentioneering.data_processors_lib import (
            AddStartEndEvents,
            AddStartEndEventsParams,
        )
        from retentioneering.preprocessing_graph import PreprocessingGraph
        from retentioneering.preprocessing_graph.nodes import EventsNode

        p = PreprocessingGraph(source_stream=self)  # type: ignore

        node = EventsNode(processor=AddStartEndEvents(params=AddStartEndEventsParams(**{})))
        p.add_node(node=node, parents=[p.root])
        result = p.combine(node)
        del p
        collect_data_performance(
            scope="add_start_end_events",
            event_name="metadata",
            called_params={},
            performance_data={},
            eventstream_index=self._eventstream_index,
            parent_eventstream_index=self._eventstream_index,
            child_eventstream_index=result._eventstream_index,
        )

        return result



from __future__ import annotations

from typing import List, Optional, Tuple

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from retentioneering.common.constants import DATETIME_UNITS
from retentioneering.utils.doc_substitution import docstrings

from ..types import EventstreamType


class SplitSessionsHelperMixin:
    @docstrings.with_indent(12)
    @time_performance(
        scope="split_sessions",
        event_name="helper",
        event_value="combine",
    )
    def split_sessions(
        self: EventstreamType,
        timeout: Optional[Tuple[float, DATETIME_UNITS]] = None,
        delimiter_events: Optional[List[str]] = None,
        delimiter_col: Optional[str] = None,
        session_col: str = "session_id",
        mark_truncated: Optional[bool] = False,
    ) -> EventstreamType:
        """
        A method of ``Eventstream`` class that creates new synthetic events in each user's path:
        ``session_start`` (or ``session_start_cropped``) and ``session_end`` (or ``session_end_cropped``).
        The created events divide users' paths on sessions.
        Also creates a new column that contains session number for each event in the input eventstream
        Session number will take the form: ``{user_id}_{session_number through one user path}``.
        The created events and column are added to the input eventstream.

        Parameters
        ----------
            %(SplitSessions.parameters)s

        Returns
        -------
        Eventstream
             Input ``eventstream`` with new synthetic events and ``session_col``.

        """
        calling_params = {
            "timeout": timeout,
            "session_col": session_col,
            "mark_truncated": mark_truncated,
            "delimiter_events": delimiter_events,
            "delimiter_col": delimiter_col,
        }
        not_hash_values = ["timeout"]

        # avoid circular import
        from retentioneering.data_processors_lib import (
            SplitSessions,
            SplitSessionsParams,
        )
        from retentioneering.preprocessing_graph import PreprocessingGraph
        from retentioneering.preprocessing_graph.nodes import EventsNode

        p = PreprocessingGraph(source_stream=self)  # type: ignore
        params = dict(
            timeout=timeout,
            delimiter_events=delimiter_events,
            delimiter_col=delimiter_col,
            session_col=session_col,
            mark_truncated=mark_truncated,
        )
        node = EventsNode(processor=SplitSessions(params=SplitSessionsParams(**params)))  # type: ignore
        p.add_node(node=node, parents=[p.root])
        result = p.combine(node)
        del p
        collect_data_performance(
            scope="split_sessions",
            event_name="metadata",
            called_params=calling_params,
            not_hash_values=not_hash_values,
            performance_data={},
            eventstream_index=self._eventstream_index,
            parent_eventstream_index=self._eventstream_index,
            child_eventstream_index=result._eventstream_index,
        )

        return result



from __future__ import annotations

from typing import Any, Literal, Optional

import pandas as pd

from retentioneering.backend.tracker import collect_data_performance, time_performance
from retentioneering.data_processor import DataProcessor
from retentioneering.eventstream.segments import (
    _calculate_segment_col,
    _get_segments_last_values,
)
from retentioneering.eventstream.types import EventstreamSchemaType
from retentioneering.params_model import ParamsModel
from retentioneering.utils.doc_substitution import docstrings
from retentioneering.utils.hash_object import hash_dataframe


class TruncatePathsParams(ParamsModel):
    """
    A class with parameters for :py:class:`.TruncatePath` class.
    """

    drop_before: Optional[str]
    drop_after: Optional[str]
    occurrence_before: Literal["first", "last"] = "first"
    occurrence_after: Literal["first", "last"] = "first"
    shift_before: int = 0
    shift_after: int = 0
    ignore_before: bool = False
    ignore_after: bool = False
    keep_synthetic: bool = False


@docstrings.get_sections(base="TruncatePath")  # type: ignore
class TruncatePaths(DataProcessor):
    """
    Leave a sub-path bounded with a given conditions. Left and right boundary conditions are associated
    with ``before`` and ``after`` argument suffixes correspondingly. If any of ``before`` and ``after``
    conditions do not meet, the path is excluded from the resulting eventstream entirely.

    Parameters
    ----------
    drop_before : str, optional
        Event name before which part of the user's path is dropped. The specified event remains in the data.
    drop_after : str, optional
        Event name after which part of the user's path is dropped. The specified event remains in the data.
    occurrence_before : {"first", "last"}, default "first"
        This parameter is necessary when the specified event occurs more than once in one user's path.

        - when set to ``first``, the part of the user path before the first event occurrence is dropped;
        - when set to ``last``, the part of the user path before the last event occurrence is dropped;
    occurrence_after : {"first", "last"}, default "first"
        The same behavior as in the 'occurrence_before', but for the other part of the user path.
    shift_before : int,  default 0
        Sets the number of steps by which the truncate point is shifted from the selected event.
        If the value is negative, then the offset occurs to the left along the timeline.
        If positive, then it occurs to the right.
    shift_after : int,  default 0
        The same behavior as in the ``shift_before``, but for the other part of the user path.
    ignore_before : bool,  default False
        If True, the resulting evenstream includes the paths that do not meet ``before`` condition.
    ignore_after : bool,  default False
        If True, the resulting evenstream includes the paths that do not meet ``after`` condition.
    keep_synthetic : bool, default False
        If True, all the synthetic events that are associated with the boundary events are kept in the output
        eventstream.

    Returns
    -------
    Eventstream
        ``Eventstream`` with the truncated sub-paths.

    Notes
    -----
    - See :doc:`Data processors user guide</user_guides/dataprocessors>` for the details.
    - If a path contains a segment synthetic event before a ``before`` cut point, the last segment value
      that occurred before the cut point is inherited. See :doc:`segments user guide</user_guides/segments_and_clusters>`.

    """

    params: TruncatePathsParams

    @time_performance(
        scope="truncate_paths",
        event_name="init",
    )
    def __init__(self, params: TruncatePathsParams):
        super().__init__(params=params)

    @time_performance(
        scope="truncate_paths",
        event_name="apply",
    )
    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        user_col = schema.user_id
        event_col = schema.event_name
        event_type_col = schema.event_type

        drop_before = self.params.drop_before
        drop_after = self.params.drop_after
        occurrence_before = self.params.occurrence_before
        occurrence_after = self.params.occurrence_after
        shift_before = self.params.shift_before
        shift_after = self.params.shift_after
        ignore_before = self.params.ignore_before
        ignore_after = self.params.ignore_after
        keep_synthetic = self.params.keep_synthetic

        params_data: list[Any] = []

        if not drop_after and not drop_before:
            raise Exception("Either drop_before or drop_after must be specified!")

        if not keep_synthetic:
            df["new_index"] = 1
            df["new_index"] = df.groupby(user_col)["new_index"].cumsum()
            index_col = "new_index"
        else:
            index_col = schema.event_index

        # tm = truncate marks
        tm = df.copy()

        if drop_before:
            before: list[str | list[str | int | None]] | None = [
                drop_before,
                ["before", occurrence_before, shift_before],
            ]
            params_data.append(before)

        if drop_after:
            after: list[str | list[str | int | None]] | None = [drop_after, ["after", occurrence_after, shift_after]]
            params_data.append(after)

        for truncate_type in params_data:
            col_mark, occurrence, shift = truncate_type[1]

            if truncate_type[0]:
                mask_events = tm[event_col] == truncate_type[0]
                tm[f"{col_mark}_mark_target"] = mask_events.astype(int)
                tm[f"{col_mark}_mark_target"] = tm.groupby([user_col, index_col])[f"{col_mark}_mark_target"].transform(
                    max
                )
                if occurrence == "last":
                    tm[f"{col_mark}_cumsum"] = tm.iloc[::-1].groupby([user_col])[f"{col_mark}_mark_target"].cumsum()
                if occurrence == "first":
                    tm[f"{col_mark}_cumsum"] = tm.groupby([user_col])[f"{col_mark}_mark_target"].cumsum()

                def count_groups(x: pd.DataFrame) -> int:
                    return x.to_frame(name=index_col).groupby(index_col).ngroup()  # type: ignore

                tm[f"{col_mark}_group_num_in_user"] = tm.groupby([user_col], group_keys=False)[index_col].transform(
                    count_groups
                )

                if occurrence == "last":
                    df_groups = (
                        tm[tm[f"{col_mark}_cumsum"] == 1]
                        .groupby([user_col])[f"{col_mark}_group_num_in_user"]
                        .max()
                        .rename(f"{col_mark}_group_centered")
                        .reset_index()
                    )
                else:
                    df_groups = (
                        tm[tm[f"{col_mark}_cumsum"] == 1]
                        .groupby([user_col])[f"{col_mark}_group_num_in_user"]
                        .min()
                        .rename(f"{col_mark}_group_centered")
                        .reset_index()
                    )

                tm = tm.merge(df_groups, how="left")
                tm[f"{col_mark}_group_centered"] = (
                    tm[f"{col_mark}_group_num_in_user"] - tm[f"{col_mark}_group_centered"] - shift
                )

        mask = pd.Series([True] * len(tm))
        if drop_before:
            if not ignore_before:
                mask &= tm["before_group_centered"] >= 0
            else:
                mask &= (tm["before_group_centered"] >= 0) | (tm["before_group_centered"].isna())

            last_segments = _get_segments_last_values(tm[tm["before_cumsum"] == 0], schema=schema)
            mask = mask | tm.index.isin(last_segments.index)

        if drop_after:
            if not ignore_after:
                mask &= tm["after_group_centered"] <= 0
            else:
                mask &= (tm["after_group_centered"] <= 0) | (tm["after_group_centered"].isna())

        if not keep_synthetic:
            df = df.drop("new_index", axis=1)

        result = df[mask]

        collect_data_performance(
            scope="truncate_paths",
            event_name="metadata",
            called_params=self.to_dict()["values"],
            not_hash_values=["occurrence_before", "occurrence_after"],
            performance_data={
                "parent": {
                    "shape": df.shape,
                    "hash": hash_dataframe(df),
                },
                "child": {
                    "shape": result.shape,
                    "hash": hash_dataframe(result),
                },
            },
        )

        return result



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TypedDict

from typing_extensions import NotRequired

from .column import Column, ColumnId
from .edges import EdgeItem, EdgesInitialState
from .nodes import NodeItem, NodesSortField, NodesSortOrder, NodeTarget
from .normalization import Normalization, NormalizationId
from .settings import SerializedSettings
from .state_changes import StateChanges
from .threshold import ThresholdValueMap
from .tracker import Tracker


class SerializedNodesState(TypedDict):
    columns: list[Column]
    items: list[NodeItem]
    normalizations: list[Normalization]
    selectedNormalizationId: NormalizationId | None
    selectedWeightsColumnId: ColumnId
    sortField: NodesSortField
    sortOrder: NodesSortOrder
    targets: list[NodeTarget]
    threshold: ThresholdValueMap
    weightsRange: ThresholdValueMap


class SerializedEdgesState(TypedDict):
    columns: list[Column]
    initialState: NotRequired[EdgesInitialState]
    items: list[EdgeItem]
    normalizations: list[Normalization]
    selectedNormalizationId: NormalizationId
    selectedWeightsColumnId: ColumnId
    threshold: ThresholdValueMap
    weightsRange: ThresholdValueMap


@dataclass
class SerializedState:
    tracker: Tracker
    nodes: SerializedNodesState
    edges: SerializedEdgesState
    settings: SerializedSettings
    recalculationChanges: StateChanges = field(default_factory=StateChanges)
    stateChanges: StateChanges = field(default_factory=StateChanges)
    restoreNodesPoints: bool | None = None
    version: str | None = None



from __future__ import annotations

import numpy as np
import pandas as pd

from retentioneering.backend.tracker import collect_data_performance, time_performance
from retentioneering.data_processor import DataProcessor
from retentioneering.eventstream.segments import (
    SEGMENT_DELIMITER,
    SEGMENT_TYPE,
    _create_segment_event,
    _extract_segment_values,
    _get_segment_mask,
)
from retentioneering.eventstream.types import (
    AddSegmentType,
    EventstreamSchemaType,
    EventstreamType,
)
from retentioneering.params_model import ParamsModel
from retentioneering.utils.doc_substitution import docstrings
from retentioneering.utils.hash_object import hash_dataframe
from retentioneering.widget.widgets import ListOfString, ReteFunction


class RenameSegmentParams(ParamsModel):
    old_name: str
    new_name: str


@docstrings.get_sections(base="RenameSegment")  # type: ignore
class RenameSegment(DataProcessor):
    """
    Rename segment for synthetic eventstream events.

    Parameters
    ----------
    old_name : str
        Old segment name to change.
    new_name : str
        New segment name to set.

    Returns
    -------
    EventstreamType
        Eventstream with renamed segment.
    """

    params: RenameSegmentParams

    @time_performance(scope="rename_segment", event_name="init")
    def __init__(self, params: RenameSegmentParams) -> None:
        super().__init__(params=params)

    @time_performance(scope="rename_segment", event_name="apply")
    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        old_name = self.params.old_name
        new_name = self.params.new_name
        hash_before = hash_dataframe(df)
        shape_before = df.shape

        mask = _get_segment_mask(df, schema, old_name)
        df.loc[mask, schema.event_name] = df.loc[mask, schema.event_name].apply(lambda x: x.replace(old_name, new_name))

        collect_data_performance(
            scope="rename_segment",
            event_name="metadata",
            called_params=self.to_dict()["values"],
            performance_data={
                "parent": {
                    "shape": shape_before,
                    "hash": hash_before,
                },
                "child": {
                    "shape": df.shape,
                    "hash": hash_dataframe(df),
                },
            },
        )
        return df



from __future__ import annotations

import itertools
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Literal

import matplotlib
import pandas as pd
import seaborn as sns

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
    tracker,
)
from retentioneering.eventstream.segments import _split_segment
from retentioneering.eventstream.types import (
    EventstreamType,
    SplitExpr,
    UserGroupsNamesType,
    UserGroupsType,
    UserListType,
)
from retentioneering.tooling.mixins.ended_events import EndedEventsMixin


@dataclass
class CenteredParams:
    event: str
    left_gap: int
    occurrence: int

    def __post_init__(self) -> None:
        if self.occurrence < 1:
            raise ValueError("Occurrence in 'centered' dictionary must be >=1")
        if self.left_gap < 1:
            raise ValueError("left_gap in 'centered' dictionary must be >=1")


class StepMatrix(EndedEventsMixin):
    """
    Step matrix is a matrix where its ``(i, j)`` element shows the frequency
    of event ``i`` occurring as ``j``-th step in user trajectories. This class
    provides methods for step matrix calculation and visualization.

    Parameters
    ----------
    eventstream : EventstreamType

    See Also
    --------
    .Eventstream.step_matrix : Call StepMatrix tool as an eventstream method.
    .StepSankey : A class for the visualization of user paths in stepwise manner using Sankey diagram.
    .CollapseLoops : Find loops and create new synthetic events in the paths of all users having such sequences.

    Notes
    -----
    See :doc:`StepMatrix user guide</user_guides/step_matrix>` for the details.
    """

    __eventstream: EventstreamType
    ENDED_EVENT = "ENDED"
    max_steps: int
    weight_col: str
    precision: int
    targets: list[str] | str | None = None
    accumulated: Literal["both", "only"] | None = None
    sorting: list | None = None
    threshold: float
    centered: CenteredParams | None = None
    groups: UserGroupsType | None = None
    group_names: UserGroupsNamesType | None = None

    _result_data: pd.DataFrame
    _result_targets: pd.DataFrame | None
    _title: str | None
    _targets_list: list[list[str]] | None

    @time_performance(
        scope="step_matrix",
        event_name="init",
    )
    def __init__(
        self,
        eventstream: EventstreamType,
    ) -> None:
        super().__init__()

        self.__eventstream = eventstream
        self.user_col = self.__eventstream.schema.user_id
        self.event_col = self.__eventstream.schema.event_name
        self.time_col = self.__eventstream.schema.event_timestamp
        self.event_index_col = self.__eventstream.schema.event_index

        self._result_data = pd.DataFrame()
        self._result_targets = None
        self._title = None
        self._targets_list = None

    def _pad_to_center(self, df_: pd.DataFrame) -> pd.DataFrame | None:
        if self.centered is None:
            return None

        center_event = self.centered.event
        occurrence = self.centered.occurrence
        window = self.centered.left_gap
        position = df_.loc[(df_[self.event_col] == center_event) & (df_["occurrence_counter"] == occurrence)][
            "event_rank"
        ].min()
        shift = position - window - 1
        df_["event_rank"] = df_["event_rank"] - shift
        return df_

    @staticmethod
    def _align_index(df1: pd.DataFrame, df2: pd.DataFrame) -> tuple[pd.DataFrame, pd.DataFrame]:
        df1 = df1.align(df2)[0].fillna(0)  # type: ignore
        df2 = df2.align(df1)[0].fillna(0)  # type: ignore
        return df1, df2  # type: ignore

    def _pad_cols(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Parameters
        ----------
        df : pd.Dataframe

        Returns
        -------
        pd.Dataframe
            With columns from 0 to ``max_steps``.
        """
        df = df.copy()
        if max(df.columns) < self.max_steps:  # type: ignore
            for col in range(max(df.columns) + 1, self.max_steps + 1):  # type: ignore
                df[col] = 0
        # add missing cols if needed:
        if min(df.columns) > 1:  # type: ignore
            for col in range(1, min(df.columns)):  # type: ignore
                df[col] = 0
        # sort cols
        return df[list(range(1, self.max_steps + 1))]

    def _step_matrix_values(
        self, data: pd.DataFrame
    ) -> tuple[pd.DataFrame, pd.DataFrame | None, str, list[list[str]] | None]:
        data = data.copy()

        # ALIGN DATA IF CENTRAL
        if self.centered is not None:
            data, fraction_title = self._center_matrix(data)
        else:
            fraction_title = ""

        # calculate step matrix elements:
        piv = self._generate_step_matrix(data)

        # ADD ROWS FOR TARGETS:
        if self.targets:
            piv_targets, targets = self._process_targets(data)
        else:
            targets, piv_targets = None, None

        return piv, piv_targets, fraction_title, targets

    def _generate_step_matrix(self, data: pd.DataFrame) -> pd.DataFrame:
        agg = data.groupby(["event_rank", self.event_col])[self.weight_col].nunique().reset_index()
        agg[self.weight_col] /= data[self.weight_col].nunique()
        agg = agg[agg["event_rank"] <= self.max_steps]
        agg.columns = ["event_rank", "event_name", "freq"]  # type: ignore
        piv = agg.pivot(index="event_name", columns="event_rank", values="freq").fillna(0)
        # add missing cols if number of events < max_steps:
        piv = self._pad_cols(piv)
        piv.columns.name = None
        piv.index.name = None
        # MAKE TERMINATED STATE ACCUMULATED:
        if self.ENDED_EVENT in piv.index:
            piv.loc[self.ENDED_EVENT] = piv.loc[self.ENDED_EVENT].cumsum().fillna(0)
        return piv

    def _process_targets(self, data: pd.DataFrame) -> tuple[pd.DataFrame | None, list[list[str]] | None]:
        if self.targets is None:
            return None, None

        # format targets to list of lists. E.g. [['a', 'b'], 'c'] -> [['a', 'b'], ['c']]
        targets = []
        if isinstance(self.targets, list):
            for t in self.targets:
                if isinstance(t, list):
                    targets.append(t)
                else:
                    targets.append([t])
        else:
            targets.append([self.targets])

        # obtain flatten list of targets. E.g. [['a', 'b'], 'c'] -> ['a', 'b', 'c']
        targets_flatten = list(itertools.chain(*targets))

        agg_targets = data.groupby(["event_rank", self.event_col])[self.time_col].count().reset_index()
        agg_targets[self.time_col] /= data[self.weight_col].nunique()
        agg_targets.columns = ["event_rank", "event_name", "freq"]  # type: ignore
        agg_targets = agg_targets[agg_targets["event_rank"] <= self.max_steps]
        piv_targets = agg_targets.pivot(index="event_name", columns="event_rank", values="freq").fillna(0)
        piv_targets = self._pad_cols(piv_targets)
        # if target is not present in dataset add zeros:
        for i in targets_flatten:
            if i not in piv_targets.index:
                piv_targets.loc[i] = 0
        piv_targets = piv_targets.loc[targets_flatten].copy()
        piv_targets.columns.name = None
        piv_targets.index.name = None
        ACC_INDEX = "ACC_"

        if self.accumulated == "only":
            piv_targets.index = map(lambda x: ACC_INDEX + x, piv_targets.index)  # type: ignore
            for i in piv_targets.index:
                piv_targets.loc[i] = piv_targets.loc[i].cumsum().fillna(0)

            # change names is targets list:
            for target in targets:
                for j, item in enumerate(target):
                    target[j] = ACC_INDEX + item
        if self.accumulated == "both":
            for i in piv_targets.index:
                piv_targets.loc[ACC_INDEX + i] = piv_targets.loc[i].cumsum().fillna(0)  # type: ignore

            # add accumulated targets to the list:
            targets_not_acc = deepcopy(targets)
            for target in targets:
                for j, item in enumerate(target):
                    target[j] = ACC_INDEX + item
            targets = targets_not_acc + targets
        return piv_targets, targets

    def _center_matrix(self, data: pd.DataFrame) -> tuple[pd.DataFrame, str]:
        if self.centered is None:
            return pd.DataFrame(), ""

        center_event = self.centered.event
        occurrence = self.centered.occurrence
        if center_event not in data[self.event_col].unique():
            error_text = 'Event "{}" from \'centered\' dict not found in the column: "{}"'.format(
                center_event, self.event_col
            )
            raise ValueError(error_text)
        # keep only users who have center_event at least N = occurrence times
        data["occurrence"] = data[self.event_col] == center_event
        data["occurrence_counter"] = data.groupby(self.weight_col)["occurrence"].cumsum() * data["occurrence"]
        users_to_keep = data[data["occurrence_counter"] == occurrence][self.weight_col].unique()
        if len(users_to_keep) == 0:
            raise ValueError(f'no records found with event "{center_event}" occurring N={occurrence} times')

        fraction_used = len(users_to_keep) / data[self.weight_col].nunique() * 100
        if fraction_used < 100:
            fraction_title = f"({fraction_used:.1f}% of total records"
        else:
            fraction_title = ""
        data = data[data[self.weight_col].isin(users_to_keep)].copy()
        data = data.groupby(self.weight_col, group_keys=True).apply(self._pad_to_center)  # type: ignore
        data = data[data["event_rank"] > 0].copy()
        return data, fraction_title

    @staticmethod
    def _sort_matrix(step_matrix: pd.DataFrame) -> pd.DataFrame:
        x = step_matrix.copy()
        order = []
        for i in x.columns:
            new_r = x[i].idxmax()  # type: ignore
            order.append(new_r)
            x = x.drop(new_r)  # type: ignore
            if x.shape[0] == 0:
                break
        order.extend(list(set(step_matrix.index) - set(order)))
        return step_matrix.loc[order]

    def _create_title(self, fraction_title: str) -> str:
        title_centered = "centered" if self.centered else ""
        title_diff = "differential " if self.groups else ""
        title_threshold = "threshold enabled)" if self.threshold > 0 else ""
        if self.threshold > 0 and fraction_title == "":
            title_technical = "("
        elif self.threshold > 0 and fraction_title != "":
            title_technical = ", "
        elif self.threshold == 0 and fraction_title != "":
            title_technical = ")"
        else:
            title_technical = ""

        if self.groups:
            if self.group_names:
                title_groups = f", {self.group_names[0]} vs. {self.group_names[1]}"
            else:
                title_groups = ", group 1 vs. group 2"
        else:
            title_groups = ""

        title = f"{title_centered} {title_diff}step matrix{title_groups}\n"
        title += f"{fraction_title}{title_technical}{title_threshold}"

        return title

    def _render_plot(
        self,
        data: pd.DataFrame,
        targets: pd.DataFrame | None,
        targets_list: list[list[str]] | None,
        title: str | None,
    ) -> Any:
        shape = len(data) + len(targets) if targets is not None else len(data), data.shape[1]
        collect_data_performance(
            scope="step_matrix",
            event_name="metadata",
            called_params={},
            performance_data={"unique_nodes": data.shape[0], "shape": shape},
            eventstream_index=self.__eventstream._eventstream_index,
        )

        n_rows = 1 + (len(targets_list) if targets_list else 0)
        n_cols = 1
        grid_specs = (
            {"wspace": 0.08, "hspace": 0.04, "height_ratios": [data.shape[0], *list(map(len, targets_list))]}
            if targets is not None and targets_list is not None
            else {}
        )
        figure, axs = sns.mpl.pyplot.subplots(
            n_rows,
            n_cols,
            sharex=True,
            figsize=(
                round(data.shape[1] * 0.7),
                round((len(data) + (len(targets) if targets is not None else 0)) * 0.6),
            ),
            gridspec_kw=grid_specs,
        )
        heatmap = sns.heatmap(
            data,
            yticklabels=data.index,
            annot=True,
            fmt=f".{self.precision}f",
            ax=axs[0] if targets is not None else axs,
            cmap="RdGy",
            center=0,
            cbar=False,
        )
        heatmap.set_title(title, fontsize=16)
        if targets is not None and targets_list is not None:
            target_cmaps = itertools.cycle(["BrBG", "PuOr", "PRGn", "RdBu"])
            for n, i in enumerate(targets_list):
                sns.heatmap(
                    targets.loc[i],
                    yticklabels=targets.loc[i].index,
                    annot=True,
                    fmt=f".{self.precision}f",
                    ax=axs[1 + n],
                    cmap=next(target_cmaps),
                    center=0,
                    vmin=targets.loc[i].values.min(),
                    vmax=targets.loc[i].values.max() or 1,
                    cbar=False,
                )

            for ax in axs:
                sns.mpl.pyplot.sca(ax)
                sns.mpl.pyplot.yticks(rotation=0)

                # add vertical lines for central step-matrix
                if self.centered is not None:
                    centered_position = self.centered.left_gap
                    ax.vlines(
                        [centered_position - 0.02, centered_position + 0.98],
                        *ax.get_ylim(),
                        colors="Black",
                        linewidth=0.7,
                    )

        else:
            sns.mpl.pyplot.sca(axs)
            sns.mpl.pyplot.yticks(rotation=0)
            # add vertical lines for central step-matrix
            if self.centered is not None:
                centered_position = self.centered.left_gap
                axs.vlines(
                    [centered_position - 0.02, centered_position + 0.98], *axs.get_ylim(), colors="Black", linewidth=0.7
                )
        return axs

    @time_performance(
        scope="step_matrix",
        event_name="fit",
    )
    def fit(
        self,
        max_steps: int = 20,
        weight_col: str | None = None,
        precision: int = 2,
        targets: list[str] | str | None = None,
        accumulated: Literal["both", "only"] | None = None,
        sorting: list | None = None,
        threshold: float = 0.01,
        centered: dict | None = None,
        groups: SplitExpr | None = None,
    ) -> None:
        """
        Calculates the step matrix internal values with the defined parameters.
        Applying ``fit`` method is necessary for the following usage
        of any visualization or descriptive ``StepMatrix`` methods.

        Parameters
        ----------
        max_steps : int, default 20
            Maximum number of steps in ``user path`` to include.
        weight_col : str, optional
            Aggregation column for edge weighting. If ``None``, specified ``user_id``
            from ``eventstream.schema`` will be used. For example, can be specified as
            ``session_id`` if ``eventstream`` has such ``custom_col``.
        precision : int, default 2
            Number of decimal digits after 0 to show as fractions in the ``heatmap``.
        targets : list of str or str, optional
            List of event names to include in the bottom of ``step_matrix`` as individual rows.
            Each specified target will have separate color-coding space for clear visualization.
            `Example: ['product_page', 'cart', 'payment']`

            If multiple targets need to be compared and plotted using the same color-coding scale,
            such targets must be combined in a sub-list.
            `Example: ['product_page', ['cart', 'payment']]`
        accumulated : {"both", "only"}, optional
            Option to include accumulated values for targets.

            - If ``None``, accumulated tartes are not shown.
            - If ``both``, show step values and accumulated values.
            - If ``only``, show targets only as accumulated.
        sorting : list of str, optional
            - If list of event names specified - lines in the heatmap will be shown in
              the passed order.
            - If ``None`` - rows will be ordered according to i`th value (first row,
              where 1st element is max; second row, where second element is max; etc)
        threshold : float, default 0.01
            Used to remove rare events. Aggregates all rows where all values are
            less than the specified threshold.
        centered : dict, optional
            Parameter used to align user paths at a specific event at a specific step.
            Has to contain three keys:
            - ``event``: str, name of event to align.
            - ``left_gap``: int, number of events to include before specified event.
            - ``occurrence`` : int which occurrence of event to align (typical 1).

            If not ``None`` - only users who have selected events with the specified
            ``occurrence`` in their paths will be included.
            ``Fraction`` of such remaining users is specified in the title of centered
            step_matrix.
            `Example: {'event': 'cart', 'left_gap': 8, 'occurrence': 1}`
        groups : tuple[list, list], tuple[str, str, str], str, optional
            Specify two groups of paths to plot differential step_matrix.
            Two separate step matrices M1 and M2 will be calculated for these groups.
            Resulting matrix is M = M1 - M2.

            - If ``tuple[list, list]``, each sub-list should contain valid path ids.
            - If ``tuple[str, str, str]``, the first str should refer to a segment name,
              the others should refer to the corresponding segment values.
            - If ``str``, it should refer to a binary (i.e. containing two segment values only) segment name.

        Notes
        -----
        During step matrix calculation an artificial ``ENDED`` event is created. If a path already
        contains ``path_end`` event (See :py:class:`.AddStartEndEvents`), it
        will be temporarily replaced with ``ENDED`` (within step matrix only). Otherwise, ``ENDED``
        event will be explicitly added to the end of each path.

        Event ``ENDED`` is cumulated so that the values in its row are summed up from
        the first step to the last. ``ENDED`` row is always placed at the last line of step matrix.
        This design guarantees that the sum of any step matrix's column is 1
        (0 for a differential step matrix).

        """
        called_params = {
            "max_steps": max_steps,
            "weight_col": weight_col,
            "precision": precision,
            "targets": targets,
            "accumulated": accumulated,
            "sorting": sorting,
            "threshold": threshold,
            "centered": centered,
            "groups": groups,
        }
        not_hash_values = ["accumulated", "centered"]

        self.max_steps = max_steps
        self.precision = precision
        self.targets = targets
        self.accumulated = accumulated
        self.sorting = sorting
        self.threshold = threshold
        self.centered = CenteredParams(**centered) if centered else None

        if groups:
            groups_, group_names = _split_segment(self.__eventstream, groups)
            self.groups = groups_
            self.group_names = group_names

        self.weight_col = weight_col or self.__eventstream.schema.user_id
        weight_col = self.weight_col or self.user_col
        data = self.__eventstream.to_dataframe()

        data = self._add_ended_events(data=data, schema=self.__eventstream.schema, weight_col=self.weight_col)
        data["event_rank"] = data.groupby(weight_col).cumcount() + 1

        # BY HERE WE NEED TO OBTAIN FINAL DIFF piv and piv_targets before sorting, thresholding and plotting:

        if self.groups:
            data_pos = data[data[weight_col].isin(self.groups[0])].copy()
            if len(data_pos) == 0:
                raise IndexError("Users from positive group are not present in dataset")
            piv_pos, piv_targets_pos, fraction_title, targets_plot = self._step_matrix_values(data=data_pos)

            data_neg = data[data[weight_col].isin(self.groups[1])].copy()
            if len(data_pos) == 0:
                raise IndexError("Users from negative group are not present in dataset")
            piv_neg, piv_targets_neg, fraction_title, targets_plot = self._step_matrix_values(data=data_neg)

            piv_pos, piv_neg = self._align_index(piv_pos, piv_neg)
            piv = piv_pos - piv_neg

            if self.targets and len(piv_targets_pos) > 0 and len(piv_targets_neg) > 0:  # type: ignore
                piv_targets_pos, piv_targets_neg = self._align_index(piv_targets_pos, piv_targets_neg)  # type: ignore
                piv_targets = piv_targets_pos - piv_targets_neg
            else:
                piv_targets = None

        else:
            piv, piv_targets, fraction_title, targets_plot = self._step_matrix_values(data=data)

        threshold_index = "THRESHOLDED_"

        if self.threshold != 0:
            # find if there are any rows to threshold:
            thresholded = piv.loc[(piv.abs() < self.threshold).all(axis=1) & (piv.index != self.ENDED_EVENT)].copy()
            if len(thresholded) > 0:
                piv = piv.loc[(piv.abs() >= self.threshold).any(axis=1) | (piv.index == self.ENDED_EVENT)].copy()
                threshold_index = f"{threshold_index}{len(thresholded)}"
                piv.loc[threshold_index] = thresholded.sum()

        if self.sorting is None:
            piv = self._sort_matrix(piv)

            keep_in_the_end = []
            keep_in_the_end.append(self.ENDED_EVENT) if (self.ENDED_EVENT in piv.index) else None
            keep_in_the_end.append(threshold_index) if (threshold_index in piv.index) else None

            events_order = [*(i for i in piv.index if i not in keep_in_the_end), *keep_in_the_end]
            piv = piv.loc[events_order]

        else:
            if {*self.sorting} != {*piv.index}:
                raise ValueError(
                    "The sorting list provided does not match the list of events. "
                    "Run with `sorting` = None to get the actual list"
                )

            piv = piv.loc[self.sorting]

        if self.centered:
            window = self.centered.left_gap
            piv.columns = [f"{int(i) - window - 1}" for i in piv.columns]  # type: ignore
            if self.targets and piv_targets is not None:
                piv_targets.columns = [f"{int(i) - window - 1}" for i in piv_targets.columns]  # type: ignore

        full_title = self._create_title(fraction_title)

        self._result_data = piv
        self._result_targets = piv_targets
        self._title = full_title
        self._targets_list = targets_plot

        shape = (
            (
                len(self._result_data) + len(self._result_targets)
                if self._result_targets is not None
                else len(self._result_data)
            ),
            self._result_data.shape[1],
        )
        collect_data_performance(
            scope="step_matrix",
            event_name="metadata",
            called_params=called_params,
            not_hash_values=not_hash_values,
            performance_data={"unique_nodes": self._result_data.shape[0], "shape": shape},
            eventstream_index=self.__eventstream._eventstream_index,
        )

    @time_performance(
        scope="step_matrix",
        event_name="plot",
    )
    def plot(self) -> Any:
        """
        Create a heatmap plot based on the calculated step matrix values.
        Should be used after :py:func:`fit`.

        Returns
        -------
        matplotlib.axes.Axes

        """
        axes = self._render_plot(self._result_data, self._result_targets, self._targets_list, self._title)
        return axes

    @property
    @time_performance(
        scope="step_matrix",
        event_name="values",
    )
    def values(self) -> tuple[pd.DataFrame, pd.DataFrame | None]:
        """
        Returns the calculated step matrix as a pd.DataFrame.
        Should be used after :py:func:`fit`.

        Returns
        -------
        tuple[pd.DataFrame, pd.DataFrame | None]
            1. Stands for the step matrix.
            2. Stands for a separate step matrix related for target events only.
        """
        return self._result_data, self._result_targets

    @property
    @time_performance(  # type: ignore
        scope="step_matrix",
        event_name="params",
    )
    def params(self) -> dict:
        """
        Returns the parameters used for the last fitting.
        Should be used after :py:func:`fit`.

        """
        return {
            "max_steps": self.max_steps,
            "weight_col": self.weight_col,
            "precision": self.precision,
            "targets": self.targets,
            "accumulated": self.accumulated,
            "sorting": self.sorting,
            "threshold": self.threshold,
            "centered": self.centered,
            "groups": self.groups,
        }



from __future__ import annotations

import pandas as pd
from pandas import DataFrame

from retentioneering.backend.tracker import collect_data_performance, time_performance
from retentioneering.data_processor import DataProcessor
from retentioneering.eventstream.types import EventstreamSchemaType
from retentioneering.params_model import ParamsModel
from retentioneering.utils.doc_substitution import docstrings
from retentioneering.utils.hash_object import hash_dataframe


class AddStartEndEventsParams(ParamsModel):
    pass


@docstrings.get_sections(base="AddStartEndEvents")  # type: ignore
class AddStartEndEvents(DataProcessor):
    """
    Create two synthetic events in each user's path: ``path_start`` and ``path_end``.
    If ``path_start`` or ``path_end`` already exists in a path, a new one is not added.

    Returns
    -------
    Eventstream
        ``Eventstream`` with new synthetic events only - two for each user:

        +----------------+----------------+----------------+
        | **event_name** | **event_type** | **timestamp**  |
        +----------------+----------------+----------------+
        | path_start     | path_start     | first_event    |
        +----------------+----------------+----------------+
        | path_end       | path_end       | last_event     |
        +----------------+----------------+----------------+

    Notes
    -----
    See :doc:`Data processors user guide</user_guides/dataprocessors>` for the details.
    """

    params: AddStartEndEventsParams

    @time_performance(scope="add_start_end_events", event_name="init")
    def __init__(self, params: AddStartEndEventsParams) -> None:
        super().__init__(params=params)

    @time_performance(
        scope="add_start_end_events",
        event_name="apply",
    )
    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        user_col = schema.user_id
        type_col = schema.event_type
        event_col = schema.event_name

        matched_events_start: DataFrame = df.groupby(user_col, as_index=False).first()  # type: ignore
        matched_events_start = matched_events_start[matched_events_start[type_col] != "path_start"]
        matched_events_start[type_col] = "path_start"
        matched_events_start[event_col] = "path_start"

        matched_events_end: DataFrame = df.groupby(user_col, as_index=False).last()  # type: ignore
        matched_events_end = matched_events_end[matched_events_end[type_col] != "path_end"]
        matched_events_end[type_col] = "path_end"
        matched_events_end[event_col] = "path_end"

        matched_events = pd.concat([matched_events_start, matched_events_end])

        result = pd.concat([df, matched_events])
        collect_data_performance(
            scope="add_start_end_events",
            event_name="metadata",
            called_params={},
            performance_data={
                "parent": {
                    "shape": df.shape,
                    "hash": hash_dataframe(df),
                },
                "child": {
                    "shape": result.shape,
                    "hash": hash_dataframe(result),
                },
            },
        )
        return result



from __future__ import annotations

import numpy as np
import pandas as pd

from retentioneering.backend.tracker import collect_data_performance, time_performance
from retentioneering.data_processor import DataProcessor
from retentioneering.eventstream.segments import (
    SEGMENT_DELIMITER,
    SEGMENT_TYPE,
    _calculate_segment_col,
    _create_segment_event,
    _get_segment_mask,
)
from retentioneering.eventstream.types import (
    AddSegmentType,
    EventstreamSchemaType,
    EventstreamType,
    SeriesWithValidator,
)
from retentioneering.params_model import ParamsModel
from retentioneering.utils.doc_substitution import docstrings
from retentioneering.utils.hash_object import hash_dataframe


class MaterializeSegmentParams(ParamsModel):
    name: str


@docstrings.get_sections(base="MaterializeSegment")  # type: ignore
class MaterializeSegment(DataProcessor):
    """
    Propagate segment synthetic events to a column.

    Parameters
    ----------
    name : str
        Name of a segment to convert

    Returns
    -------
    Eventstream
        Eventstream with added column.
    """

    params: MaterializeSegmentParams

    @time_performance(scope="materialize_segment", event_name="init")
    def __init__(self, params: MaterializeSegmentParams) -> None:
        super().__init__(params=params)

    @time_performance(scope="materialize_segment", event_name="apply")
    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        name = self.params.name
        hash_before = hash_dataframe(df)
        shape_before = df.shape

        segment_col = _calculate_segment_col(df, schema, name)
        df[name] = segment_col

        collect_data_performance(
            scope="rename_segment",
            event_name="metadata",
            called_params=self.to_dict()["values"],
            performance_data={
                "parent": {
                    "shape": shape_before,
                    "hash": hash_before,
                },
                "child": {
                    "shape": df.shape,
                    "hash": hash_dataframe(df),
                },
            },
        )

        return df



from __future__ import annotations

from typing import Tuple

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from retentioneering.common.constants import DATETIME_UNITS
from retentioneering.utils.doc_substitution import docstrings

from ..types import EventstreamType


class DropPathsHelperMixin:
    @docstrings.with_indent(12)
    @time_performance(
        scope="drop_paths",
        event_name="helper",
        event_value="combine",
    )
    def drop_paths(
        self: EventstreamType, min_steps: int | None = None, min_time: Tuple[float, DATETIME_UNITS] | None = None
    ) -> EventstreamType:
        """
        A method of ``Eventstream`` class that deletes users' paths that are shorter than the specified
        number of events or cut_off.

        Parameters
        ----------
            %(DropPaths.parameters)s

        Returns
        -------
        Eventstream
             Input ``eventstream`` without the deleted short users' paths.


        """
        calling_params = {
            "min_steps": min_steps,
            "min_time": min_time,
        }
        not_hash_values = ["min_time"]

        # avoid circular import
        from retentioneering.data_processors_lib import DropPaths, DropPathsParams
        from retentioneering.preprocessing_graph import PreprocessingGraph
        from retentioneering.preprocessing_graph.nodes import EventsNode

        p = PreprocessingGraph(source_stream=self)  # type: ignore

        node = EventsNode(
            processor=DropPaths(params=DropPathsParams(min_steps=min_steps, min_time=min_time))  # type: ignore
        )
        p.add_node(node=node, parents=[p.root])
        result = p.combine(node)
        del p
        collect_data_performance(
            scope="drop_paths",
            event_name="metadata",
            called_params=calling_params,
            not_hash_values=not_hash_values,
            performance_data={},
            eventstream_index=self._eventstream_index,
            parent_eventstream_index=self._eventstream_index,
            child_eventstream_index=result._eventstream_index,
        )

        return result
